package dk.nsp.idsas.dao;

import dk.nsp.idsas.dao.entity.BlurringEntity;
import dk.nsp.idsas.service.IdsasServiceException;
import dk.nsp.idsas.util.ErrorMessages;
import dk.nsp.idsas.util.FaultCode;
import org.springframework.dao.DataAccessException;
import org.apache.log4j.Logger;
import org.springframework.jdbc.core.BeanPropertyRowMapper;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;

import javax.sql.DataSource;
import java.time.*;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static dk.nsp.idsas.dao.DbUtils.toTimestamp;
import java.util.*;
import java.util.stream.Collectors;

public class BlurringDao {
    private static final Logger logger = Logger.getLogger(BlurringDao.class);

    private final NamedParameterJdbcTemplate jdbcTemplate;
    private final BeanPropertyRowMapper<BlurringEntity> rowMapper;

    public BlurringDao(DataSource dataSource) {
        jdbcTemplate = new NamedParameterJdbcTemplate(dataSource);
        rowMapper = BeanPropertyRowMapper.newInstance(BlurringEntity.class);
    }

    public void insert(BlurringEntity blurringEntity) throws IdsasServiceException {
        String query = "INSERT INTO blurring(patient_id, patient_id_type, organisation_id, organisation_id_type, expiry_date)" +
                "           VALUES (:patient_id, :patient_id_type, :organisation_id, :organisation_id_type, :expiry_date)";
        HashMap<String, Object> parameters = new HashMap<>();
        parameters.put("patient_id", blurringEntity.getPatientId());
        parameters.put("patient_id_type", blurringEntity.getPatientIdType());
        parameters.put("organisation_id", blurringEntity.getOrganisationId());
        parameters.put("organisation_id_type", blurringEntity.getOrganisationIdType());
        parameters.put("expiry_date", toTimestamp(blurringEntity.getExpiryDate()));

        try {
            jdbcTemplate.update(query, parameters);
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.DAO_ACCESS_ERROR, false, e);
        }
    }

    public List<BlurringEntity> getActiveBlurringsByHash(String hashedPatientId, String patientIdType) throws IdsasServiceException {
        // Find de seneste sløringer, per patient/organisation, der ikke er udløbet
        String query = "SELECT b1.organisation_id, b1.organisation_id_type\n" +
                "FROM blurring b1\n" +
                "LEFT JOIN blurring b2 on (\n" +
                "  b1.patient_id_hash = b2.patient_id_hash AND\n" +
                "  b1.patient_id_type = b2.patient_id_type AND\n" +
                "  b1.organisation_id = b2.organisation_id AND\n" +
                "  b1.organisation_id_type = b2.organisation_id_type AND\n" +
                "  b1.created_time < b2.created_time)\n" +
                "WHERE b2.id is null\n" +
                "AND b1.patient_id_hash = :patient_id_hash\n" +
                "AND b1.patient_id_type = :patient_id_type\n" +
                "AND b1.expiry_date > :expiry_date;";

        HashMap<String, Object> parameters = new HashMap<>();
        parameters.put("patient_id_hash", hashedPatientId);
        parameters.put("patient_id_type", patientIdType);
        parameters.put("expiry_date", toTimestamp(Instant.now()));

        try {
            return jdbcTemplate.query(query, parameters, rowMapper);
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.DAO_ACCESS_ERROR, false, e);
        }
    }

    public List<BlurringEntity> getActiveBlurrings(String patientId, String patientIdType) throws IdsasServiceException {
        // Find de seneste sløringer, per patient/organisation, der ikke er udløbet
        String query = "SELECT b1.organisation_id, b1.organisation_id_type\n" +
                "FROM blurring b1\n" +
                "LEFT JOIN blurring b2 on (\n" +
                "  b1.patient_id = b2.patient_id AND\n" +
                "  b1.patient_id_type = b2.patient_id_type AND\n" +
                "  b1.organisation_id = b2.organisation_id AND\n" +
                "  b1.organisation_id_type = b2.organisation_id_type AND\n" +
                "  b1.created_time < b2.created_time)\n" +
                "WHERE b2.id is null\n" +
                "AND b1.patient_id = :patient_id\n" +
                "AND b1.patient_id_type = :patient_id_type\n" +
                "AND b1.expiry_date > NOW();";

        HashMap<String, Object> parameters = new HashMap<>();
        parameters.put("patient_id", patientId);
        parameters.put("patient_id_type", patientIdType);
        try {
            return jdbcTemplate.query(query, parameters, rowMapper);
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.DAO_ACCESS_ERROR, false, e);
        }
    }

    public int getInactiveBlurringsCount(int maxRetentionYears) throws IdsasServiceException {
        final String query =
                "SELECT COUNT(*) FROM blurring " +
                "WHERE expiry_date <= :retention_time ";

        Map<String, Object> paramMap = new HashMap<>();
        paramMap.put("retention_time", toTimestamp(LocalDateTime.now(ZoneOffset.UTC).minusYears(maxRetentionYears).toInstant(ZoneOffset.UTC)));

        try {
            Integer count = jdbcTemplate.queryForObject(query, paramMap, Integer.class);
            return count == null ? 0 : count;
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.DAO_ACCESS_ERROR, false, e);
        }
    }

    public int deleteRowsInBatch(int batchSize, int maxRetentionYears) throws IdsasServiceException {
        final String deleteQuery =
                "DELETE FROM blurring " +
                "      WHERE expiry_date <= :retention_time " +
                "      LIMIT :limit";

        Map<String, Object> paramMap = new HashMap<>();
        paramMap.put("limit", batchSize);
        paramMap.put("retention_time", toTimestamp(LocalDateTime.now(ZoneOffset.UTC).minusYears(maxRetentionYears).toInstant(ZoneOffset.UTC)));

        try {
            return jdbcTemplate.update(deleteQuery, paramMap);
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.DAO_ACCESS_ERROR, false, e);
        }
    }

    public List<String> getAllBlurrings() throws IdsasServiceException {
        String query = "SELECT DISTINCT patient_id FROM blurring";

        try {
            List<String> list  = jdbcTemplate.queryForList(query, new HashMap<>(), String.class);
            return list.stream()
                    .filter(Objects::nonNull)
                    .collect(Collectors.toList());
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.DAO_ACCESS_ERROR, false, e);
        }
    }

    public void deleteBlurringsForDeceased(List<String> deceasedToBeDeleted) throws IdsasServiceException {
        final String deleteQuery =
                "DELETE FROM blurring " +
                " WHERE patient_id in (:patient_id_list)";

        Map<String, Object> paramMap = new HashMap<>();
        paramMap.put("patient_id_list", deceasedToBeDeleted);

        try {
            int count = jdbcTemplate.update(deleteQuery, paramMap);
            logger.debug("Deleted " + count + " rows for deceased people");
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.DAO_ACCESS_ERROR, false, e);
        }
    }

}
