package dk.nsp.idsas.dao;

import dk.nsp.idsas.dao.entity.OrgBlurringEntity;
import dk.nsp.idsas.service.IdsasServiceException;
import dk.nsp.idsas.util.ErrorMessages;
import dk.nsp.idsas.util.FaultCode;
import org.springframework.dao.DataAccessException;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;

import javax.sql.DataSource;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.time.LocalDateTime;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeFormatterBuilder;
import java.util.HashMap;
import java.util.List;

import static dk.nsp.idsas.dao.DbUtils.toTimestamp;


public class OrgBlurringDao {
    private final NamedParameterJdbcTemplate jdbcTemplate;
    private final DateTimeFormatter formatter;

    public OrgBlurringDao(DataSource dataSource) {
        this.jdbcTemplate = new NamedParameterJdbcTemplate(dataSource);
        this.formatter = new DateTimeFormatterBuilder().append(DateTimeFormatter.ISO_OFFSET_DATE_TIME).toFormatter();
    }

    public void insert(OrgBlurringEntity orgBlurringEntity) throws IdsasServiceException {
        String query = "INSERT INTO organisation_blurring(organisation_id, organisation_id_type, creator_cvr, valid_from, valid_to, created_time)" +
                " VALUES (:organisation_id, :organisation_id_type, :creator_cvr, :valid_from, :valid_to, :created_time)";
        HashMap<String, Object> parameters = new HashMap<>();
        parameters.put("organisation_id", orgBlurringEntity.getOrganisationId());
        parameters.put("organisation_id_type", orgBlurringEntity.getOrganisationIdType());
        parameters.put("creator_cvr", orgBlurringEntity.getCreatorCvr());
        parameters.put("created_time", formatter.format(OffsetDateTime.now()));
        parameters.put("valid_from",  toTimestamp(orgBlurringEntity.getValidFrom()));
        if (orgBlurringEntity.getValidTo() != null) {
            parameters.put("valid_to", toTimestamp(orgBlurringEntity.getValidTo()));
        } else {
            parameters.put("valid_to", null);
        }

        try {
            jdbcTemplate.update(query, parameters);
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.DAO_ACCESS_ERROR, false, e);
        }
    }

    public List<OrgBlurringEntity> getActiveOrgBlurrings() throws IdsasServiceException {
        String query = "SELECT organisation_id, organisation_id_type, creator_cvr\n" +
                "FROM organisation_blurring\n" +
                "WHERE (valid_to > :now OR valid_to IS NULL)\n" +
                "AND valid_from < :now;";

        HashMap<String, Object> parameters = new HashMap<>();
        parameters.put("now", LocalDateTime.now().toString());

        try {
            return jdbcTemplate.query(query, parameters, (rs, rowNum) -> mapOrgBlurringEntityRow(rs));
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.DAO_ACCESS_ERROR, false, e);
        }
    }

    public List<OrgBlurringEntity> getActiveOrgBlurrings(String organizationId, String organizationIdType) throws IdsasServiceException {
        String query = "SELECT organisation_id, organisation_id_type, creator_cvr\n" +
                "FROM organisation_blurring\n" +
                "WHERE organisation_id = :organisation_id\n" +
                "AND organisation_id_type = :organisation_id_type\n" +
                "AND (valid_to > :now OR valid_to IS NULL)\n" +
                "AND valid_from < :now;";

        HashMap<String, Object> parameters = new HashMap<>();
        parameters.put("organisation_id", organizationId);
        parameters.put("organisation_id_type", organizationIdType);
        parameters.put("now", toTimestamp(LocalDateTime.now(ZoneOffset.UTC).toInstant(ZoneOffset.UTC)));

        try {
            return jdbcTemplate.query(query, parameters, (rs, rowNum) -> mapOrgBlurringEntityRow(rs));
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.DAO_ACCESS_ERROR, false, e);
        }
    }

    public List<OrgBlurringEntity> getActiveOrgBlurrings(String creatorCvr) throws IdsasServiceException {
        String query = "SELECT organisation_id, organisation_id_type, creator_cvr\n" +
                "FROM organisation_blurring\n" +
                "WHERE (valid_to > :now OR valid_to IS NULL)\n" +
                "AND valid_from < :now\n" +
                "AND creator_cvr = :creator_cvr\n" +
                ";";

        HashMap<String, Object> parameters = new HashMap<>();
        parameters.put("now", toTimestamp(LocalDateTime.now(ZoneOffset.UTC).toInstant(ZoneOffset.UTC)));
        parameters.put("creator_cvr", creatorCvr);

        try {
            return jdbcTemplate.query(query, parameters, (rs, rowNum) -> mapOrgBlurringEntityRow(rs));
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.DAO_ACCESS_ERROR, false, e);
        }
    }

    public boolean orgBlurringExists(String organizationId, String organizationIdType) throws IdsasServiceException {
        String query = "SELECT EXISTS(\n" +
                "SELECT 1 FROM organisation_blurring\n" +
                "WHERE organisation_id = :organisation_id\n" +
                "AND organisation_id_type = :organisation_id_type\n" +
                "AND (valid_to > :now OR valid_to IS NULL)\n" +
                "AND valid_from < :now\n" +
                ");";

        HashMap<String, Object> parameters = new HashMap<>();
        parameters.put("organisation_id", organizationId);
        parameters.put("organisation_id_type", organizationIdType);
        parameters.put("now", toTimestamp(LocalDateTime.now(ZoneOffset.UTC).toInstant(ZoneOffset.UTC)));
        try {
            return Boolean.TRUE.equals(jdbcTemplate.queryForObject(query, parameters, Boolean.class));
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.DAO_ACCESS_ERROR, false, e);
        }
    }

    public void removeOrgBlurring(String organizationId, String organizationIdType) throws IdsasServiceException {
        String query = "UPDATE organisation_blurring\n" +
                "SET valid_to = :now\n" +
                "WHERE organisation_id = :organisation_id\n" +
                "AND organisation_id_type = :organisation_id_type\n" +
                "AND valid_to IS NULL\n";

        HashMap<String, Object> parameters = new HashMap<>();
        parameters.put("organisation_id", organizationId);
        parameters.put("organisation_id_type", organizationIdType);
        parameters.put("now", toTimestamp(LocalDateTime.now(ZoneOffset.UTC).toInstant(ZoneOffset.UTC)));

        try {
            jdbcTemplate.update(query, parameters);
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.DAO_ACCESS_ERROR, false, e);
        }
    }

    private static OrgBlurringEntity mapOrgBlurringEntityRow(ResultSet rs) throws SQLException {
        OrgBlurringEntity orgBlurringEntity = new OrgBlurringEntity();
        orgBlurringEntity.setOrganisationId(rs.getString("organisation_id"));
        orgBlurringEntity.setOrganisationIdType(rs.getString("organisation_id_type"));
        orgBlurringEntity.setCreatorCvr(rs.getString("creator_cvr"));
        return orgBlurringEntity;
    }
}
