package dk.nsp.idsas.dao;

import dk.nsp.idsas.dao.entity.OrgBlurringEntity;
import dk.nsp.idsas.service.IdsasServiceException;
import org.junit.Before;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;

import javax.sql.DataSource;
import java.time.*;
import java.util.List;

import static org.junit.Assert.*;

public class OrgBlurringDaoTest extends AbstractDaoTest {
    @Autowired
    DataSource dataSource;

    private OrgBlurringDao orgBlurringDao;

    private final OrgBlurringEntity TEST_BLURRING_1 =
            new OrgBlurringEntity("1111", "SOR", "12345678",
                    instantFrom(2022, 1, 1, 0, 0), instantFrom(3000, 1, 1, 0, 0));

    private final OrgBlurringEntity TEST_BLURRING_2 =
            new OrgBlurringEntity("2222", "CVR", "12345678",
                    instantFrom(2022, 1, 1, 0, 0), instantFrom(3000, 1, 1, 0, 0));

    private final OrgBlurringEntity TEST_BLURRING_6 =
            new OrgBlurringEntity("6666", "SOR", "22221111",
                    instantFrom(2022, 1, 1, 0, 0), instantFrom(2500, 1, 1, 0, 0));

    private final OrgBlurringEntity TEST_BLURRING_8 =
            new OrgBlurringEntity("8888", "SOR", "88888888",
                    instantFrom(2022, 1, 1, 0, 0), instantFrom(3000, 1, 1, 0, 0));

    private final OrgBlurringEntity TEST_BLURRING_EXPIRED =
            new OrgBlurringEntity("5555", "SOR", "11112222",
                    instantFrom(2020, 1, 1, 0, 0), instantFrom(2021, 1, 1, 0, 0));

    private final OrgBlurringEntity TEST_BLURRING_NO_EXPIRATION =
            new OrgBlurringEntity("1122", "SOR", "33456789",
                    instantFrom(2022, 1, 1, 0, 0), null);

    @Before
    public void setUp() {
        orgBlurringDao = new OrgBlurringDao(dataSource);
    }

    @Test
    public void insert_WhenCalled_InsertsBlurring() throws IdsasServiceException {
        // Act
        orgBlurringDao.insert(TEST_BLURRING_2);

        // Assert
        List<OrgBlurringEntity> blurrings = orgBlurringDao.getActiveOrgBlurrings(TEST_BLURRING_2.getCreatorCvr());
        assertEquals(2, blurrings.size());
    }

    @Test
    public void insert_WhenCalled_InsertsBlurringWithCorrectValues() throws IdsasServiceException {
        // Arrange
        List<OrgBlurringEntity> blurrings = orgBlurringDao.getActiveOrgBlurrings(TEST_BLURRING_8.getCreatorCvr());
        int beforeSize = blurrings.size();
        orgBlurringDao.insert(TEST_BLURRING_8);

        // Assert
        blurrings = orgBlurringDao.getActiveOrgBlurrings(TEST_BLURRING_8.getCreatorCvr());
        assertEquals(beforeSize + 1, blurrings.size());
        OrgBlurringEntity returnedBlurring = blurrings.get(0);
        assertEquals(TEST_BLURRING_8.getOrganisationId(), returnedBlurring.getOrganisationId());
        assertEquals(TEST_BLURRING_8.getOrganisationIdType(), returnedBlurring.getOrganisationIdType());
    }

    @Test
    public void getAllOrgBlurrings_WhenOneMatchingActiveBlurringIsFound_ReturnsBlurring() throws IdsasServiceException {
        // Act
        List<OrgBlurringEntity> blurrings = orgBlurringDao.getActiveOrgBlurrings(TEST_BLURRING_1.getCreatorCvr());

        // Assert
        assertEquals(1, blurrings.size());
        OrgBlurringEntity returnedBlurring = blurrings.get(0);
        assertEquals(TEST_BLURRING_1.getOrganisationId(), returnedBlurring.getOrganisationId());
        assertEquals(TEST_BLURRING_1.getOrganisationIdType(), returnedBlurring.getOrganisationIdType());
    }

    @Test
    public void getAllOrgBlurrings_WhenOneActiveWithoutExpirationIsFound_ReturnsBlurring() throws IdsasServiceException {
        // Act
        List<OrgBlurringEntity> blurrings = orgBlurringDao.getActiveOrgBlurrings(TEST_BLURRING_1.getCreatorCvr());

        // Assert
        assertEquals(1, blurrings.size());
        OrgBlurringEntity returnedBlurring = blurrings.get(0);
        assertEquals(TEST_BLURRING_1.getOrganisationId(), returnedBlurring.getOrganisationId());
        assertEquals(TEST_BLURRING_1.getOrganisationIdType(), returnedBlurring.getOrganisationIdType());
    }

    @Test
    public void getAllOrgBlurrings_WhenNoMatchingBlurringIsFound_ReturnsEmptyList() throws IdsasServiceException {
        // Act
        List<OrgBlurringEntity> blurrings = orgBlurringDao.getActiveOrgBlurrings("does not exist");

        // Assert
        assertEquals(0, blurrings.size());
    }

    @Test
    public void getAllOrgBlurrings_WhenOnlyMatchesExpiredBlurring_DoesNotReturnBlurring() throws IdsasServiceException {
        // Act
        List<OrgBlurringEntity> blurrings = orgBlurringDao.getActiveOrgBlurrings(TEST_BLURRING_EXPIRED.getCreatorCvr());

        // Assert
        assertEquals(0, blurrings.size());
    }

    @Test
    public void getAllOrgBlurrings_WhenTwoMatchingActiveBlurringIsFound_ReturnsBlurring() throws IdsasServiceException {
        // Act
        List<OrgBlurringEntity> blurrings = orgBlurringDao.getActiveOrgBlurrings(TEST_BLURRING_6.getCreatorCvr());

        // Assert
        assertEquals(2, blurrings.size());
        OrgBlurringEntity returnedBlurring = blurrings.get(0);
        assertEquals(TEST_BLURRING_6.getOrganisationId(), returnedBlurring.getOrganisationId());
        assertEquals(TEST_BLURRING_6.getOrganisationIdType(), returnedBlurring.getOrganisationIdType());
    }

    @Test
    public void getActiveOrgBlurrings_WhenTwoMatchingActiveBlurringIsFound_ReturnsBlurring() throws IdsasServiceException {
        // Act
        List<OrgBlurringEntity> blurrings = orgBlurringDao.getActiveOrgBlurrings("1111", "SOR");

        // Assert
        assertEquals(2, blurrings.size());
        OrgBlurringEntity returnedBlurring = blurrings.get(0);
        assertEquals(TEST_BLURRING_1.getOrganisationId(), returnedBlurring.getOrganisationId());
        assertEquals(TEST_BLURRING_1.getOrganisationIdType(), returnedBlurring.getOrganisationIdType());
    }

    @Test
    public void getActiveOrgBlurrings_WhenOneMatchingActiveBlurringIsFound_ReturnsBlurring() throws IdsasServiceException {
        // Act
        List<OrgBlurringEntity> blurrings = orgBlurringDao.getActiveOrgBlurrings("1122", "SOR");

        // Assert
        assertEquals(1, blurrings.size());
        OrgBlurringEntity returnedBlurring = blurrings.get(0);
        assertEquals(TEST_BLURRING_NO_EXPIRATION.getOrganisationId(), returnedBlurring.getOrganisationId());
        assertEquals(TEST_BLURRING_NO_EXPIRATION.getOrganisationIdType(), returnedBlurring.getOrganisationIdType());
    }

    @Test
    public void getActiveOrgBlurrings_WhenNoMatchingActiveBlurringIsFound_ReturnsEmptyList() throws IdsasServiceException {
        // Act
        List<OrgBlurringEntity> blurrings = orgBlurringDao.getActiveOrgBlurrings("does not exist", "SOR");

        // Assert
        assertEquals(0, blurrings.size());
    }

    @Test
    public void getAllActiveOrgBlurrings_returnsThem() throws IdsasServiceException {
        // Act
        List<OrgBlurringEntity> blurrings = orgBlurringDao.getActiveOrgBlurrings();

        // Assert
        assertEquals(8, blurrings.size());
    }

    @Test
    public void orgBlurringExists_WhenBlurringExists_ReturnsTrue() throws IdsasServiceException {
        // Act
        boolean exists = orgBlurringDao.orgBlurringExists("1111", "SOR");

        // Assert
        assertTrue(exists);
    }

    @Test
    public void orgBlurringExists_WhenBlurringDoesNotExist_ReturnsFalse() throws IdsasServiceException {
        // Act
        boolean exists = orgBlurringDao.orgBlurringExists("does not exist", "SOR");

        // Assert
        assertFalse(exists);
    }

    @Test
    public void removeOrgBlurring_RemovesExistingBlurring_ItsNoLongerActive() throws IdsasServiceException {
        // Act
        List<OrgBlurringEntity> before = orgBlurringDao.getActiveOrgBlurrings("3333", "SOR");
        orgBlurringDao.removeOrgBlurring("3333", "SOR");

        // Assert
        List<OrgBlurringEntity> after = orgBlurringDao.getActiveOrgBlurrings("3333", "SOR");
        assertEquals(1, before.size());
        assertEquals(0, after.size());

    }

    private static Instant instantFrom(int year, int month, int dayOfMonth, int hour, int minute) {
        return LocalDateTime.of(year, month, dayOfMonth, hour, minute).atZone(ZoneId.systemDefault()).toInstant();
    }
}
