package dk.nsp.idsas.integrations.personinformation;

import com.fasterxml.jackson.core.type.TypeReference;
import dk.nsp.idsas.integrations.ServiceErrorCollector;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.time.Duration;
import java.time.Instant;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.stream.Collectors;

public class PersonInformationClient {
    private static final Logger logger = Logger.getLogger(PersonInformationClient.class);
    private static final String APPLICATION_JSON = "application/json";
    private static final String CONTENT_TYPE = "Content-Type";

    private final ObjectMapper objectMapper = new ObjectMapper();
    private final ServiceErrorCollector serviceErrorCollector;
    private final HttpClient httpClient;
    private final String url;

    public PersonInformationClient(HttpClient httpClient, String url, ServiceErrorCollector serviceErrorCollector) {
        this.httpClient = httpClient;
        this.url = url;
        this.serviceErrorCollector = serviceErrorCollector;
    }

    public PersonInformationClient(HttpClient httpClient, String url, Duration errorCountDuration) {
        this(httpClient, url, new ServiceErrorCollector(errorCountDuration));
    }

    public int getErrorCount() {
        return serviceErrorCollector.getErrorCount();
    }

    public boolean personExists(String cpr) throws PersonInformationClientException {
        // Eksempel på request:
        // http://test1-cnsp.ekstern-test.nspop.dk:8080/stamdata-personinformation/person/1212121212

        HttpUriRequest cprRequest = new HttpGet(String.format("%s/person/%s", url, cpr));
        return invokePersonExists(cprRequest);
    }

    public List<String> getDeceased(List<String> listOfCpr, Instant deceasedSince) throws PersonInformationClientException {
        // Eksempel på request:
        // http://test1-cnsp.ekstern-test.nspop.dk:8080/stamdata-personinformation/v1/person/deceased?cpr=0304784567&dato=2023-07-21T17:32:28

        String cprParams = listOfCpr.stream().map(cpr -> "cpr=" + cpr).collect(Collectors.joining("&"));
        HttpUriRequest request = new HttpGet(String.format("%s/person/deceased?%s&dato=%s", url, cprParams, DateTimeFormatter.ISO_INSTANT.format(deceasedSince)));
        return invokeGetDeceased(request);
    }

    private <T> T invoke(HttpUriRequest request, ResponseFunction<HttpResponse, T> responseHandler, String errorMessage) throws PersonInformationClientException {
        try {
            HttpResponse response = httpClient.execute(request);
            T result = responseHandler.apply(response);
            serviceErrorCollector.registerServiceSuccess();
            return result;
        } catch (IOException e) {
            logger.error(e.getMessage(), e);
            serviceErrorCollector.registerServiceError();
            throw new PersonInformationClientException(errorMessage, e);
        } catch (PersonInformationClientException e) {
            logger.error(e.getMessage(), e);
            serviceErrorCollector.registerServiceError();
            throw e;
        }
    }

    @FunctionalInterface
    private interface ResponseFunction<T, R> {
        R apply(T t) throws PersonInformationClientException;
    }

    private boolean invokePersonExists(HttpUriRequest request) throws PersonInformationClientException {
        return invoke(request, response -> {
            validateResponse(response);
            Map<String, Object> result = parseJsonResponseAsMap(response);
            return result.get("cpr") != null;
        }, "Fejl i kaldet til 'personExists' på personinformation.");
    }

    private List<String> invokeGetDeceased(HttpUriRequest request) throws PersonInformationClientException {
        return invoke(request, response -> {
            validateResponse(response);
            List<Map<String, Object>> responseJson = parseJsonResponseAsList(response);
            return responseJson.stream()
                    .map(map -> map.get("cpr").toString())
                    .collect(Collectors.toList());
        }, "Fejl i kaldet til 'deceased' på personinformation.");
    }

    private void validateResponse(HttpResponse response) throws PersonInformationClientException {
        int statusCode = response.getStatusLine().getStatusCode();
        if (statusCode == HttpStatus.SC_OK || statusCode == HttpStatus.SC_NOT_FOUND) {
            boolean contentTypeOk = contentTypeOk(response);
            if (!contentTypeOk) {
                logger.error(String.format("Fejl i retursvar fra PersonInformation (%d) forkert contenttype", statusCode));
                throw new PersonInformationClientException("Fejl i retursvar fra PersonInformation");
            }
        } else {
            throw new PersonInformationClientException(String.format("Kaldet til personinformation fejlede med statuskode: %d", statusCode));
        }
    }

    private List<Map<String, Object>> parseJsonResponseAsList(HttpResponse response) throws PersonInformationClientException {
        try {
            return objectMapper.readValue(response.getEntity().getContent(), new TypeReference<List<Map<String, Object>>>() {
            });
        } catch (UnsupportedOperationException | IOException e) {
            logger.error("Kunne ikke parse json til List<Map<String, Object>> type", e);
            throw new PersonInformationClientException("Fejl i fortolkning af svar fra PersonInformation", e);
        }
    }

    private Map<String, Object> parseJsonResponseAsMap(HttpResponse response) throws PersonInformationClientException {
        try {
            return objectMapper.readValue(response.getEntity().getContent(), new TypeReference<Map<String, Object>>() {
            });
        } catch (UnsupportedOperationException | IOException e) {
            logger.error("Kunne ikke parse json til Map type", e);
            throw new PersonInformationClientException("Fejl i fortolkning af svar fra PersonInformation", e);
        }
    }

    private boolean contentTypeOk(HttpResponse response) {
        Header headers = response.getFirstHeader(CONTENT_TYPE);
        return headers != null && headers.getValue().equalsIgnoreCase(APPLICATION_JSON);
    }
}
