package dk.nsp.idsas.integrations.personinformation;

import dk.nsp.idsas.integrations.ServiceErrorCollector;

import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpUriRequest;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

import java.io.IOException;
import java.time.Instant;
import java.util.Arrays;
import java.util.List;

import static dk.nsp.idsas.integrations.util.TestUtil.createHttpResponse;
import static org.junit.Assert.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.verify;

public class PersonInformationClientTest {

    private HttpClient httpClient;
    private final String url = "test-url";
    private ServiceErrorCollector serviceErrorCollector;
    private PersonInformationClient personInformationClient;

    @Before
    public void setUp() {
        httpClient = Mockito.mock(HttpClient.class);
        serviceErrorCollector = Mockito.mock(ServiceErrorCollector.class);
        personInformationClient = new PersonInformationClient(httpClient, url, serviceErrorCollector);
    }

    @Test
    public void personExists_WhenServiceReturnsValidResponse_ReturnsTrueAndRegistersServiceSuccess() throws PersonInformationClientException, IOException {
        HttpResponse httpResponse = createHttpResponse(200, "application/json", "{\"cpr\": \"1234\"}");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));
        String cpr = "1234";

        boolean personExists = personInformationClient.personExists(cpr);

        assertTrue(personExists);
        String expectedPath = url + "/person/" + cpr;
        verify(httpClient).execute(
                Mockito.argThat(request -> expectedPath.equals(request.getURI().getPath())));
        verify(serviceErrorCollector).registerServiceSuccess();
    }

    @Test
    public void personExists_WhenClientThrowsIOException_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        Mockito.doThrow(IOException.class).when(httpClient).execute(any(HttpUriRequest.class));

        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personExists("1234"));

        assertEquals("Fejl i kaldet til 'personExists' på personinformation.", e.getMessage());
        verify(serviceErrorCollector).registerServiceError();
    }

    @Test
    public void personExists_WhenServiceReturnsResponse200WithContentTypeText_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        HttpResponse httpResponse = createHttpResponse(200, "text/plain", "{\"cpr\": \"1234\"}");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personExists("1234"));

        assertEquals("Fejl i retursvar fra PersonInformation", e.getMessage());
        verify(serviceErrorCollector).registerServiceError();
    }

    @Test
    public void personExists_WhenServiceReturnsResponse200WithInvalidJson_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        HttpResponse httpResponse = createHttpResponse(200, "application/json", "invalid body");
        doAnswer(arg -> httpResponse)
                .when(httpClient).execute(any(HttpUriRequest.class));

        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personExists("1234"));

        assertEquals("Fejl i fortolkning af svar fra PersonInformation", e.getMessage());
        verify(serviceErrorCollector).registerServiceError();
    }

    @Test
    public void personExists_WhenServiceReturnsResponse200WithoutCprEntry_ReturnsFalse() throws IOException, PersonInformationClientException {
        HttpResponse httpResponse = createHttpResponse(200, "application/json", "{\"test\": \"1234\"}");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        boolean result = personInformationClient.personExists("1234");

        assertFalse(result);
    }

    @Test
    public void personExists_WhenServiceReturnsResponse404WithContentTypeText_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        HttpResponse httpResponse = createHttpResponse(404, "text/plain", "{\"cpr\": \"1234\"}");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personExists("1234"));

        assertEquals("Fejl i retursvar fra PersonInformation", e.getMessage());
        verify(serviceErrorCollector).registerServiceError();
    }

    @Test
    public void personExists_WhenServiceReturnsResponse404WithInvalidJson_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        HttpResponse httpResponse = createHttpResponse(404, "application/json", "invalid body");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personExists("1234"));

        assertEquals("Fejl i fortolkning af svar fra PersonInformation", e.getMessage());
        verify(serviceErrorCollector).registerServiceError();
    }

    @Test
    public void personExists_WhenServiceReturnsResponse404WithValidJson_ReturnsFalseAndRegistersServiceSuccess() throws PersonInformationClientException, IOException {
        HttpResponse httpResponse = createHttpResponse(404, "application/json", "{\"test\": \"test\"}");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        boolean personExists = personInformationClient.personExists("1234");

        assertFalse(personExists);
        verify(serviceErrorCollector).registerServiceSuccess();
    }

    @Test
    public void personExists_WhenServiceReturnsResponse400_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        HttpResponse httpResponse = createHttpResponse(400, "application/json", "{\"test\": \"test\"}");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personExists("1234"));

        assertEquals("Kaldet til personinformation fejlede med statuskode: 400", e.getMessage());
        verify(serviceErrorCollector).registerServiceError();
    }

    @Test
    public void deceasedExists_WhenServiceReturnsValidResponse_ReturnsTrueAndRegistersServiceSuccess() throws PersonInformationClientException, IOException {
        HttpResponse httpResponse = createHttpResponse(200, "application/json", "[{\"cpr\": \"2345\"},{\"cpr\": \"3456\"}]");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        List<String> deceasedCprs = personInformationClient.getDeceased(Arrays.asList("1234", "2345", "3456"), Instant.now());

        assertEquals(2, deceasedCprs.size());
        verify(serviceErrorCollector).registerServiceSuccess();
    }

    @Test
    public void deceasedExists_WhenServiceReturnsValidResponseWithEmptyList_ReturnsEmptyListAndRegistersServiceSuccess() throws PersonInformationClientException, IOException {
        HttpResponse httpResponse = createHttpResponse(200, "application/json", "[]");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        List<String> deceasedCprs = personInformationClient.getDeceased(Arrays.asList("1234", "2345", "3456"), Instant.now());

        assertEquals(0, deceasedCprs.size());
        verify(serviceErrorCollector).registerServiceSuccess();
    }

    @Test
    public void deceasedExists_WhenServiceReturnsValidResponseWithInvalidJson_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        HttpResponse httpResponse = createHttpResponse(200, "application/json", "invalid body");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.getDeceased(Arrays.asList("1234", "2345", "3456"), Instant.now()));

        assertEquals("Fejl i fortolkning af svar fra PersonInformation", e.getMessage());
        verify(serviceErrorCollector).registerServiceError();
    }
}