package dk.nsp.idsas.operations.setup;

import dk.nsp.idsas.dao.BlurringDao;
import dk.nsp.idsas.dao.OrgBlurringDao;
import dk.nsp.idsas.dao.SaltDao;
import dk.nsp.idsas.http.PoolingHttpClientFactory;
import dk.nsp.idsas.integrations.personinformation.PersonInformationClient;
import dk.nsp.idsas.service.CleanupDeceasedService;
import dk.nsp.idsas.kafka.KafkaUtil;
import dk.nsp.idsas.service.CleanupService;
import dk.nsp.idsas.kafka.KafkaConsumerService;
import dk.nsp.idsas.service.SaltService;
import dk.nsp.idsas.util.health.CleanupDeceasedHealthIndicator;
import dk.nsp.idsas.util.health.CleanupHealthIndicator;
import dk.nsp.idsas.util.health.KafkaConsumerHealthIndicator;
import dk.sds.nsp.kafka.clients.consumer.NSPKafkaConsumer;
import dk.nsp.idsas.util.health.PersonInformationHealthIndicator;
import org.apache.http.client.HttpClient;
import org.apache.kafka.clients.consumer.Consumer;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.PropertySource;
import org.springframework.context.annotation.PropertySources;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.security.SecureRandom;
import java.time.Duration;
import java.time.Period;
import java.util.Collections;
import java.util.Properties;

@Configuration
@PropertySources({
        @PropertySource("classpath:/META-INF/maven/dk.nsp.idsas/idsas-operations-web/pom.properties"),
        @PropertySource(value = "idsas-operations.properties", encoding = "ISO-8859-1")
        // Bemærkning: I Java 8 er properties altid i ISO-8859-1, men fremadrettet skulle de gerne kunne være UTF-8.
})
public class WebSetup {
    private static final Logger logger = Logger.getLogger(WebSetup.class);

    @Value("${idsas.cleanup.max.duration:PT15S}")
    private String cleanupMaxDuration;

    @Value("${idsas.cleanup.max.retention:PT5Y}")
    private String cleanupMaxRetention;

    @Value("${idsas.cleanup.batch.size:10000}")
    private int cleanupBatchSize;

    @Value("${personinformation.url}")
    private String personInformationUrl;

    @Value("${personinformation.errorcount.duration:PT10M}")
    private String personInformationErrorCountDuration;

    @Value("${idsas.httpclient.pooling.totalconnections:200}")
    private int maxTotalConnections;

    @Value("${idsas.httpclient.pooling.maxconnections.pr.route:20}")
    private int defaultMaxConnectionsPerRoute;

    @Value("${idsas.cleanup.deceased.period:PT1Y}")
    private String cleanupDeceasedPeriod;

    @Value("${idsas.cleanup.deceased.batch.size:10000}")
    private int cleanupDeceasedBatchSize;

    @Value("${kafka.topic}")
    private String kafkaTopic;

    @Value("${kafka.poll.timeout:1000}")
    private long kafkaPollTimeout;

    @Value("${personinformation.error.tolerance:0}")
    private int personInformationErrorTolerance;

    @Bean
    public SaltService saltService(SaltDao saltDao) {
        return new SaltService(saltDao, null, new SecureRandom(), new String[]{});
    }

    @Bean
    public CleanupService cleanupService(BlurringDao blurringDao) {
        return new CleanupService(blurringDao, Duration.parse(cleanupMaxDuration), cleanupBatchSize, Period.parse(cleanupMaxRetention).getYears());
    }

    @Bean
    public PersonInformationHealthIndicator personInformationHealthIndicator(PersonInformationClient personInformationClient) {
        return new PersonInformationHealthIndicator(personInformationClient, personInformationErrorTolerance);
    }

    @Bean
    public CleanupHealthIndicator cleanupHealthIndicator(CleanupService cleanupService) {
        return new CleanupHealthIndicator(cleanupService);
    }

    @Bean
    public CleanupDeceasedHealthIndicator cleanupDeceasedHealthIndicator(CleanupDeceasedService cleanupDeceasedService) {
        return new CleanupDeceasedHealthIndicator(cleanupDeceasedService);
    }

    @Bean
    public PoolingHttpClientFactory poolingHttpClientFactory() {
        return new PoolingHttpClientFactory(maxTotalConnections, defaultMaxConnectionsPerRoute);
    }

    @Bean
    public PersonInformationClient personInformationClient(PoolingHttpClientFactory poolingHttpClientFactory) {
        HttpClient httpClient = poolingHttpClientFactory.getClient();
        return new PersonInformationClient(httpClient, personInformationUrl, Duration.parse(personInformationErrorCountDuration));
    }

    @Bean
    public CleanupDeceasedService cleanupDeceasedService(BlurringDao blurringDao, PersonInformationClient personInformationClient) {
        return new CleanupDeceasedService(blurringDao,
                personInformationClient,
                Duration.parse(cleanupMaxDuration),
                cleanupDeceasedBatchSize,
                Period.parse(cleanupDeceasedPeriod));
    }

    @Bean
    public KafkaConsumerHealthIndicator kafkaConsumerHealthIndicator(KafkaConsumerService kafkaConsumerService) {
        return new KafkaConsumerHealthIndicator(kafkaConsumerService);
    }

    @Bean
    public Properties kafkaProperties() throws IOException {
        final String propertiesFile = "idsas-operations.properties";
        final String prefix = "kafka.consumer.";

        // Read the properties
        Properties properties = new Properties();
        try (InputStream inputStream = getClass().getClassLoader().getResourceAsStream(propertiesFile)) {
            if (inputStream != null) {
                properties.load(inputStream);
            } else {
                throw new FileNotFoundException(String.format("Property file '%s' not found in the classpath", propertiesFile));
            }
        }

        return KafkaUtil.filterKafkaProperties(properties, prefix);
    }

    @Bean
    public Consumer<String, String> kafkaConsumer(Properties kafkaProperties) {
        Consumer<String, String> consumer = new NSPKafkaConsumer<>(kafkaProperties);
        consumer.subscribe(Collections.singletonList(kafkaTopic));
        return consumer;
    }

    @Bean
    public KafkaConsumerService kafkaConsumerService(Consumer<String, String> kafkaConsumer,
                                                     BlurringDao blurringDao, OrgBlurringDao orgBlurringDao) {
        return new KafkaConsumerService(kafkaConsumer, kafkaPollTimeout, blurringDao, orgBlurringDao);
    }
}
