package dk.nsp.idsas.kafka;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;

import dk.nsp.idsas.dao.BlurringDao;
import dk.nsp.idsas.dao.OrgBlurringDao;
import dk.nsp.idsas.dao.entity.BlurringEntity;
import dk.nsp.idsas.dao.entity.OrgBlurringEntity;
import dk.nsp.idsas.dto.KafkaMsgType;
import dk.nsp.idsas.service.IdsasServiceException;
import dk.nsp.idsas.util.FaultCode;
import dk.nsp.idsas.util.health.KafkaConsumeStatus;
import org.apache.kafka.clients.consumer.Consumer;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.common.errors.WakeupException;
import org.apache.log4j.Logger;

import java.time.OffsetDateTime;
import java.time.format.DateTimeFormatter;
import java.util.concurrent.atomic.AtomicReference;

import static dk.nsp.idsas.dto.KafkaMsgType.*;

public class KafkaConsumerService {
    private static final Logger logger = Logger.getLogger(KafkaConsumerService.class);
    private static final Logger auditLogger = Logger.getLogger("consumerAuditLogger");

    private final Consumer<String, String> consumer;
    private final long pollTimeout;
    private final AtomicReference<KafkaConsumeStatus> status = new AtomicReference<>();
    private final BlurringDao blurringDao;
    private final OrgBlurringDao orgBlurringDao;
    private final ObjectMapper mapper = new ObjectMapper();

    public KafkaConsumerService(Consumer<String, String> consumer, long pollTimeout,
                                BlurringDao blurringDao, OrgBlurringDao orgBlurringDao) {
        this.consumer = consumer;
        this.pollTimeout = pollTimeout;
        this.blurringDao = blurringDao;
        this.orgBlurringDao = orgBlurringDao;

        status.set(KafkaConsumeStatus.NOT_STARTED);

        // Så Instants konverteres til ISO-datoer
        mapper.disable(SerializationFeature.WRITE_DATES_AS_TIMESTAMPS);
        mapper.registerModule(new JavaTimeModule());
    }

    public synchronized void consume() throws IdsasServiceException {
        try {
            status.set(KafkaConsumeStatus.RUNNING);
            ConsumerRecords<String, String> consumerRecords = consumer.poll(pollTimeout);
            final int count = consumerRecords.count();
            logger.info("KafkaConsumerService Job");
            logger.info(String.format("status=%s", status.get()));
            logger.info(String.format("received-records=%d", count));

            if (consumerRecords.isEmpty()) {
                status.set(KafkaConsumeStatus.FINISHED);
                return;
            }

            for (ConsumerRecord<String, String> consumerRecord : consumerRecords) {
                String jsonString = consumerRecord.value();
                KafkaMessage message = mapper.readValue(jsonString, KafkaMessage.class);
                logger.debug(String.format("Received message of type: %s, key: %s", message.getType(), message.getKey()));

                KafkaMsgType msgType = message.getType();
                switch (msgType) {
                    case CREATE_BLURRING_ENTITY: {
                        logger.trace("Processing CREATE_BLURRING_ENTITY message");
                        BlurringEntity entity = mapper.readValue(message.getMessage(), BlurringEntity.class);
                        blurringDao.insert(entity);
                        auditLog(message.getKey(), CREATE_BLURRING_ENTITY.getValue(), entity);
                        break;
                    }
                    case CREATE_ORG_BLURRING_ENTITY: {
                        logger.trace("Processing CREATE_ORG_BLURRING_ENTITY message");
                        OrgBlurringEntity entity = mapper.readValue(message.getMessage(), OrgBlurringEntity.class);

                        if (orgBlurringDao.orgBlurringExists(entity.getOrganisationId(), entity.getOrganisationIdType())) {
                            // Den her situation opstår hvis den samme sløring oprettes flere gange, inden den første er blevet
                            // persisteret i databasen (fordi den har siddet i Kafka-køen).
                            logger.warn(String.format("Organisation blurring already exists and cannot be inserted again. Organisation id: %s and organisation id type: %s",
                                    entity.getOrganisationId(), entity.getOrganisationIdType()));
                            return;
                        } else {
                            orgBlurringDao.insert(entity);
                            auditLog(message.getKey(), CREATE_ORG_BLURRING_ENTITY.getValue(), entity);
                        }

                        break;
                    }
                    case REMOVE_ORG_BLURRING_ENTITY: {
                        logger.trace("Processing REMOVE_ORG_BLURRING_ENTITY message");
                        OrgBlurringEntity entity = mapper.readValue(message.getMessage(), OrgBlurringEntity.class);
                        orgBlurringDao.removeOrgBlurring(entity.getOrganisationId(), entity.getOrganisationIdType());
                        auditLog(message.getKey(), REMOVE_ORG_BLURRING_ENTITY.getValue(), entity);

                        break;
                    }
                    default:
                        logger.error("Unknown Kafka message type: " + message.getType());
                        throw new IdsasServiceException("Unknown Kafka message type: " + message.getType(), FaultCode.CREATE_BLURRING_ERROR, false);
                }
            }

            consumer.commitSync();
            status.set(KafkaConsumeStatus.FINISHED);
        } catch (WakeupException e) {
            status.set(KafkaConsumeStatus.ERROR);
            throw new IdsasServiceException("Wakeup called on Kafka consumer", FaultCode.CREATE_BLURRING_ERROR, false, e);
        } catch (JsonProcessingException e) {
            status.set(KafkaConsumeStatus.ERROR);
            throw new IdsasServiceException("Error parsing JSON message", FaultCode.CREATE_BLURRING_ERROR, false, e);
        } finally {
            logger.info(String.format("status=%s", status.get()));
        }
    }

    public void stopConsume() {
        consumer.wakeup();
    }

    public KafkaConsumeStatus getStatus() {
        return status.get();
    }

    protected void auditLog(String key, String type, Object message) {
        ObjectNode jsonObject = mapper.createObjectNode();

        jsonObject.put("time", DateTimeFormatter.ISO_OFFSET_DATE_TIME.format(OffsetDateTime.now()));
        jsonObject.put("key", key);
        jsonObject.put("type", type);
        jsonObject.set("message", mapper.valueToTree(message));

        auditLogger.info(jsonObject);
    }
}
