package dk.nsp.idsas.kafka;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;
import dk.nsp.idsas.dao.entity.BlurringEntity;
import dk.nsp.idsas.dao.entity.OrgBlurringEntity;
import dk.nsp.idsas.dto.KafkaMsgType;
import dk.nsp.idsas.service.IdsasServiceException;
import dk.nsp.idsas.util.ErrorMessages;
import dk.nsp.idsas.util.FaultCode;
import org.apache.kafka.clients.producer.Producer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.clients.producer.RecordMetadata;
import org.apache.log4j.Logger;

import java.time.OffsetDateTime;
import java.time.format.DateTimeFormatter;
import java.util.UUID;
import java.util.concurrent.Future;

public class KafkaProducerService {
    public static final String KAFKA_CLIENT_ID_KEY = "client.id";

    private static final Logger logger = Logger.getLogger(KafkaProducerService.class);
    private static final Logger auditLogger = Logger.getLogger("producerAuditLogger");

    private final String topic;
    private final Producer<String, String> kafkaProducer;

    private final ObjectMapper mapper = new ObjectMapper();

    public KafkaProducerService(String topic, Producer<String, String> kafkaProducer) {
        this.topic = topic;
        this.kafkaProducer = kafkaProducer;

        // Så vi kan få serialiseret vores Instants som ISO datoer
        mapper.disable(SerializationFeature.WRITE_DATES_AS_TIMESTAMPS);
        mapper.registerModule(new JavaTimeModule());
    }

    public void sendCreateBlurringMessage(BlurringEntity entity) throws IdsasServiceException {
        sendMessage(entity, KafkaMsgType.CREATE_BLURRING_ENTITY);
    }

    public void sendCreateOrgBlurringMessage(OrgBlurringEntity entity) throws IdsasServiceException {
        sendMessage(entity, KafkaMsgType.CREATE_ORG_BLURRING_ENTITY);
    }

    public void removeOrgBlurringMessage(OrgBlurringEntity entity) throws IdsasServiceException {
        sendMessage(entity, KafkaMsgType.REMOVE_ORG_BLURRING_ENTITY);
    }

    private void sendMessage(Object o, KafkaMsgType type) throws IdsasServiceException {
        try {
            String key = UUID.randomUUID().toString();
            KafkaMessage message = new KafkaMessage(key, type, mapper.writeValueAsString(o));
            String value = mapper.writeValueAsString(message);
            ProducerRecord<String, String> producerRecord = new ProducerRecord<>(topic, key, value);

            logger.debug(String.format("Sending Kafka message. Key: %s", key));
            Future<RecordMetadata> recordMetadataFuture = kafkaProducer.send(producerRecord);
            RecordMetadata recordMetadata = recordMetadataFuture.get();
            if (recordMetadata.topic() == null) {
                throw new Exception(ErrorMessages.unknownKafkaTopic());
            }

            auditLog(message.getKey(), type.getValue(), o);
        } catch (Exception e) {
            String msg = "Fejl ved afsendelse af Kafka besked.";
            throw new IdsasServiceException(ErrorMessages.blurringError(msg), FaultCode.CREATE_BLURRING_ERROR, false, e);
        }
    }

    protected void auditLog(String key, String type, Object message) {
        ObjectNode jsonObject = mapper.createObjectNode();

        jsonObject.put("time", DateTimeFormatter.ISO_OFFSET_DATE_TIME.format(OffsetDateTime.now()));
        jsonObject.put("key", key);
        jsonObject.put("type", type);
        jsonObject.set("message", mapper.valueToTree(message));

        auditLogger.info(jsonObject);
    }
}
