package dk.nsp.idsas.service;

import dk.nsp.idsas.dao.WhitelistDao;
import dk.nsp.idsas.dao.entity.BlurringEntity;
import dk.nsp.idsas.dto.OrgIdType;
import dk.nsp.idsas.dto.PersonIdType;
import dk.nsp.idsas.integrations.personinformation.PersonInformationClient;
import dk.nsp.idsas.integrations.personinformation.PersonInformationClientException;
import dk.nsp.idsas.util.AuditLogger;
import dk.nsp.idsas.util.ErrorMessages;
import dk.nsp.idsas.util.WhitelistType;
import dk.nsp.idsas.util.FaultCode;

import java.time.ZonedDateTime;

public class BlurringService {
    private final WhitelistDao whitelistDao;
    private final AuditLogger auditLogger;
    private final PersonInformationClient personInformationClient;
    private final BlurringStorage<BlurringEntity> createStorage;
    private final int expiryMaxDays;

    public BlurringService(WhitelistDao whitelistDao, AuditLogger auditLogger, PersonInformationClient personInformationClient, BlurringStorage<BlurringEntity> createStorage, int expiryMaxDays) {
        this.whitelistDao = whitelistDao;
        this.auditLogger = auditLogger;
        this.personInformationClient = personInformationClient;
        this.expiryMaxDays = expiryMaxDays;
        this.createStorage = createStorage;
    }

    public void createBlurring(String organizationId, OrgIdType organizationIdType, String patientId, PersonIdType patientIdType, ZonedDateTime endDateTime) throws IdsasServiceException {

        // Man kan kun være whitelisted med orgIdType "CVR"
        if (!organizationIdType.equals(OrgIdType.CVR)) {
            throw new IdsasServiceException(ErrorMessages.whitelistingActorIdType(organizationIdType.name(), OrgIdType.CVR.name()), FaultCode.CREATE_ORG_BLURRING_ERROR, true);
        }

        // Tjek at organizationId er whitelisted til at kalde createBlurring
        boolean isWhiteListed = whitelistDao.isWhitelisted(organizationId, WhitelistType.BLURRING);
        if (!isWhiteListed) {
            throw new IdsasServiceException(ErrorMessages.cvrWhitelistError(organizationId, WhitelistType.BLURRING), FaultCode.CREATE_BLURRING_ERROR, true);
        }

        // Tjek at patientId er gyldig (via "personinformation")
        if (patientIdType.equals(PersonIdType.CPR)) {
            boolean patientIdIsValid = false;
            try {
                patientIdIsValid = personInformationClient.personExists(patientId);
                if (!patientIdIsValid) {
                    throw new IdsasServiceException(ErrorMessages.cprIsNotValidError(), FaultCode.CREATE_BLURRING_ERROR, true);
                }
            } catch (PersonInformationClientException e) {
                throw new IdsasServiceException(e.getMessage(), FaultCode.CREATE_BLURRING_ERROR, false, e);
            }
        }

        // Tjek at sløringen ikke ligger mere end 'expiryMaxDays' fremme i tiden
        if (endDateTime.isAfter(ZonedDateTime.now().plusDays(expiryMaxDays))) {
            throw new IdsasServiceException(ErrorMessages.maxExpiryDateExceeded(expiryMaxDays), FaultCode.CREATE_BLURRING_ERROR, true);
        }

        // Entity oprettes og gemmes
        BlurringEntity entity = new BlurringEntity(
                patientId,
                patientIdType.name(),
                organizationId,
                organizationIdType.name(),
                endDateTime.toInstant());

        // Blurring oprettes via valgte storage strategi
        createStorage.apply(entity);

        // Audit logning
        auditLogger.logCreateBlurring(patientId, patientIdType.name(), organizationId, organizationIdType.name(), entity.getExpiryDate());
    }

}
