package dk.nsp.idsas.service;

import dk.nsp.idsas.dao.BlurringDao;
import dk.nsp.idsas.integrations.personinformation.PersonInformationClient;
import dk.nsp.idsas.util.health.CleanupJobStatus;
import org.apache.log4j.Logger;

import java.time.Duration;
import java.time.Instant;
import java.time.Period;
import java.time.ZonedDateTime;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.atomic.AtomicReference;

public class CleanupDeceasedService {
    private static final Logger logger = Logger.getLogger(CleanupDeceasedService.class);

    private final BlurringDao blurringDao;
    private final PersonInformationClient personInformationClient;
    private final Period deceasedPeriod;
    private final Duration maxDuration;
    private final int batchSize;
    private final AtomicReference<CleanupJobStatus> status = new AtomicReference<>();

    private List<String> peopleWorkQueue;

    public CleanupDeceasedService(BlurringDao blurringDao,
                                  PersonInformationClient personInformationClient,
                                  Duration maxDuration,
                                  int batchSize,
                                  Period deceasedPeriod) {
        this.blurringDao = blurringDao;
        this.personInformationClient = personInformationClient;
        this.maxDuration = maxDuration;
        this.batchSize = batchSize;
        this.deceasedPeriod = deceasedPeriod;
        this.status.set(CleanupJobStatus.notStarted());

        peopleWorkQueue = new LinkedList<>();
    }

    public CleanupJobStatus getStatus() {
        return status.get();
    }

    public void cleanUpDeceasedBlurrings() {
        logger.info("Starting cleanup deceased job");

        int deceasedDeletedCount = 0;

        try {
            loadWorkQueue();
            logger.info("deceasedWorkQueue=" +  peopleWorkQueue.size());

            if (peopleWorkQueue.isEmpty()) {
                logger.info("No deceased persons to delete.");
                status.set(new CleanupJobStatus(0, 0, CleanupJobStatus.ExecutionStatus.FINISHED));
            } else {
                // Start the clock
                long startMillis = System.currentTimeMillis();
                Instant latestDeceasedTime = ZonedDateTime.now().minus(deceasedPeriod).toInstant();

                while (belowTimeLimit(startMillis) && !peopleWorkQueue.isEmpty()) {

                    // Find batch of people
                    List<String> personBatch = getNextBatchInWorkQueue();
                    // Find those that are deceased
                    List<String> deceasedToBeDeleted = personInformationClient.getDeceased(personBatch, latestDeceasedTime);

                    if (!deceasedToBeDeleted.isEmpty()) {
                        blurringDao.deleteBlurringsForDeceased(deceasedToBeDeleted);
                        deceasedDeletedCount += deceasedToBeDeleted.size();
                    }
                }
            }
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
        }

        writeToLog(deceasedDeletedCount);
    }

    private void writeToLog(int deceasedDeletedCount) {
        StringBuilder logInfo = new StringBuilder();
        status.set(new CleanupJobStatus(deceasedDeletedCount, peopleWorkQueue.size(),
                peopleWorkQueue.isEmpty() ?
                        CleanupJobStatus.ExecutionStatus.FINISHED : CleanupJobStatus.ExecutionStatus.NOT_FINISHED));
        logInfo.append(String.format("deleted=%s ", deceasedDeletedCount));
        logInfo.append(String.format("remaining=%s ", peopleWorkQueue.size()));
        logInfo.append(String.format("status=%s", status.get().finished));

        logger.info(logInfo);
    }

    private synchronized void loadWorkQueue() throws IdsasServiceException {
        if (peopleWorkQueue.isEmpty()) {
            peopleWorkQueue = blurringDao.getAllBlurrings();
        }
    }

    private synchronized List<String> getNextBatchInWorkQueue() {
        List<String> batch = new LinkedList<>();
        // Tag hele køen, hvis den er mindre end batchSze
        if (peopleWorkQueue.size() <= batchSize) {
            batch = new LinkedList<>(peopleWorkQueue);
            peopleWorkQueue = new LinkedList<>();
        } else if (!peopleWorkQueue.isEmpty()) {
            // Ellers tag op til batchSize elementer fra køen
            batch = peopleWorkQueue.subList(0, batchSize);
            // ... og fjern dem fra køen
            peopleWorkQueue = peopleWorkQueue.subList(batchSize, peopleWorkQueue.size());
        }
        return batch;
    }

    protected boolean belowTimeLimit(long startMillis) {
        return ((System.currentTimeMillis() - startMillis) < maxDuration.toMillis());
    }
}
