package dk.nsp.idsas.service;

import dk.nsp.idsas.dao.BlurringDao;
import dk.nsp.idsas.dao.entity.BlurringEntity;
import dk.nsp.idsas.dto.OrgIdType;
import dk.nsp.idsas.dto.OrganizationIdDto;
import dk.nsp.idsas.dto.PersonIdType;
import dk.nsp.idsas.security.IdsasActor;
import dk.nsp.idsas.util.AuditLogger;
import dk.nsp.idsas.util.ErrorMessages;
import dk.nsp.idsas.util.FaultCode;
import dk.nsp.idsas.util.HashingUtil;
import org.apache.log4j.Logger;
import org.springframework.dao.DataAccessException;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

public class LookupService {
    private static final Logger logger = Logger.getLogger(LookupService.class);

    private final BlurringDao blurringDao;
    private final AuditLogger auditLogger;
    private final List<String> allowedClientSystemNames;
    private final String patientIdSaltPrefix;
    private final String patientIdSaltPostfix;
    private final LookupStrategy lookupStrategy;

    public LookupService(BlurringDao blurringDao, String patientIdSaltPrefix, String patientIdSaltPostfix,
                         AuditLogger auditLogger, String[] allowedClientSystemNames, LookupStrategy lookupStrategy) {
        this.blurringDao = blurringDao;
        this.auditLogger = auditLogger;
        this.allowedClientSystemNames = Arrays.asList(allowedClientSystemNames);
        this.patientIdSaltPrefix = patientIdSaltPrefix;
        this.patientIdSaltPostfix = patientIdSaltPostfix;
        this.lookupStrategy = lookupStrategy;

        logger.info("LookupService initialized with lookup strategy: " + lookupStrategy);
    }
    
    public List<OrganizationIdDto> getActiveBlurrings(IdsasActor actor, String patientId, PersonIdType patientIdType) throws IdsasServiceException {
        // Validering af, at actor er en systembruger
        if (!actor.getActorType().equals(IdsasActor.IdsasActorType.SYSTEM)) {
            throw new IdsasServiceException(ErrorMessages.userTypeNotSupported(actor.getActorType().name()), FaultCode.LOOKUP_ERROR, true);
        }

        // Validering af, at systembrugeren har et clientName med den korrekte værdi
        if (!allowedClientSystemNames.contains(actor.getClientSystemName())) {
            String names = String.join(", ", allowedClientSystemNames);
            throw new IdsasServiceException(ErrorMessages.clientSystemNameError(
                    actor.getClientSystemName(), names), FaultCode.LOOKUP_ERROR, true);
        }

        // Til de hentede entiteter
        List<BlurringEntity> entityList = new ArrayList<>();

        // Hent de hashede, hvis det er valgt
        if (lookupStrategy.equals(LookupStrategy.HASH) || lookupStrategy.equals(LookupStrategy.BOTH)) {
            try {
                String hashedPatientId = HashingUtil.createPatientIdHash(patientId, patientIdSaltPrefix, patientIdSaltPostfix);
                entityList.addAll(blurringDao.getActiveBlurringsByHash(hashedPatientId, patientIdType.name()));
            } catch (DataAccessException e) {
                throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.LOOKUP_ERROR, false, e);
            }
        }

        // Hent via cpr-nummer, hvis det er valgt
        if (lookupStrategy.equals(LookupStrategy.CPR) || lookupStrategy.equals(LookupStrategy.BOTH)) {
            try {
                entityList.addAll(blurringDao.getActiveBlurrings(patientId, patientIdType.name()));
            } catch (DataAccessException e) {
                throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.LOOKUP_ERROR, false, e);
            }
        }

        // Mappe entiteter til liste af id'ere
        List<OrganizationIdDto> orgIdList = entityList.stream().map(
            e -> new OrganizationIdDto(e.getOrganisationId(), OrgIdType.valueOf(e.getOrganisationIdType()))
        ).collect(Collectors.toList());

        // Audit logning
        auditLogger.logGetBlurredOrganisations(patientId, patientIdType.name(), !entityList.isEmpty());

        return orgIdList;
    }
}

