package dk.nsp.idsas.service;

import dk.nsp.idsas.dao.OrgBlurringDao;
import dk.nsp.idsas.dao.WhitelistDao;
import dk.nsp.idsas.dao.entity.OrgBlurringEntity;
import dk.nsp.idsas.dto.OrgIdType;
import dk.nsp.idsas.dto.OrganizationIdDto;
import dk.nsp.idsas.integrations.sores.SoresClient;
import dk.nsp.idsas.integrations.sores.SoresClientException;
import dk.nsp.idsas.security.IdsasActor;
import dk.nsp.idsas.util.ErrorMessages;
import dk.nsp.idsas.util.FaultCode;
import dk.nsp.idsas.util.WhitelistType;
import org.apache.log4j.Logger;

import java.time.Instant;
import java.util.*;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

public class OrgBlurringService {
    private static final Logger logger = Logger.getLogger(OrgBlurringService.class);

    private final WhitelistDao whitelistDao;
    private final OrgBlurringDao orgBlurringDao;
    private final SoresClient soresClient;
    private final Set<String> allowedClientSystemNames;
    private final BlurringStorage<OrgBlurringEntity> createStorage;
    private final BlurringStorage<OrgBlurringEntity> removeStorage;

    public OrgBlurringService(WhitelistDao whitelistDao, OrgBlurringDao orgBlurringDao, SoresClient soresClient,
                              String[] allowedClientSystemNames,
                              BlurringStorage<OrgBlurringEntity> createStorage,
                              BlurringStorage<OrgBlurringEntity> removeStorage) {
        this.whitelistDao = whitelistDao;
        this.orgBlurringDao = orgBlurringDao;
        this.soresClient = soresClient;
        this.allowedClientSystemNames = new HashSet<>(Arrays.asList(allowedClientSystemNames));
        this.createStorage = createStorage;
        this.removeStorage = removeStorage;
    }

    public OrgBlurringService(WhitelistDao whitelistDao, OrgBlurringDao orgBlurringDao, SoresClient soresClient,
                              String[] allowedClientSystemNames) {
        this.whitelistDao = whitelistDao;
        this.orgBlurringDao = orgBlurringDao;
        this.soresClient = soresClient;
        this.allowedClientSystemNames = new HashSet<>(Arrays.asList(allowedClientSystemNames));

        // Bruges ikke i lookup-servicen
        this.createStorage = null;
        this.removeStorage = null;
    }

    /**
     * Opretter en ny afdelingssløring.<br/>
     * Understøttede organisationer i createOrgBlurring er SOR og SHAK. SOR-koder valideres på om de eksisterer via kald til SORES.
     * SHAK koder kontrolleres ikke.<br/>
     */
    public void createOrgBlurring(String creatorCvr, IdsasActor.IdsasActorType actorType, String organizationId, OrgIdType organizationIdType) throws IdsasServiceException {
        logger.trace("Create org blurring");

        // Tjek at det er en SYSTEM-bruger der kalder
        if (!actorType.equals(IdsasActor.IdsasActorType.SYSTEM)) {
            throw new IdsasServiceException(ErrorMessages.userTypeNotSupported(actorType.name()),
                    FaultCode.CREATE_ORG_BLURRING_ERROR, true);
        }

        // Tjek at creatorCvr er whitelisted til at kalde createBlurring / createOrgBlurring
        boolean isWhiteListed = whitelistDao.isWhitelisted(creatorCvr, WhitelistType.BLURRING);
        if (!isWhiteListed) {
            throw new IdsasServiceException(ErrorMessages.cvrWhitelistError(creatorCvr, WhitelistType.BLURRING),
                    FaultCode.CREATE_ORG_BLURRING_ERROR, true);
        }

        // Tjek at organizationId er enten SOR eller SHAK
        if (!organizationIdType.equals(OrgIdType.SOR) && !organizationIdType.equals(OrgIdType.SHAK)) {
            throw new IdsasServiceException(ErrorMessages.orgIdTypeNotSupported(organizationIdType.name()),
                    FaultCode.CREATE_ORG_BLURRING_ERROR, true);
        }

        // Hvis der registreres med SOR-kode, skal vi tjekke at SOR-koden hører under opretterens CVR-nummer
        if (organizationIdType.equals(OrgIdType.SOR)) {
            try {
                String cvr = soresClient.getCvrNumberBySor(organizationId);
                if (cvr == null || !cvr.equals(creatorCvr)) {
                    throw new IdsasServiceException(ErrorMessages.sorDoesNotBelongToCvr(organizationId, creatorCvr), FaultCode.CREATE_ORG_BLURRING_ERROR, true);
                }
            } catch (SoresClientException e) {
                throw new IdsasServiceException(ErrorMessages.sorClientError(), FaultCode.CREATE_ORG_BLURRING_ERROR, false, e);
            }
        }

        // Hvis der registreres med SHAK-kode, skal vi tjekke, at den kun indeholder tal og store bogstaver
        if (organizationIdType.equals(OrgIdType.SHAK) && !Pattern.matches("[A-Z0-9]+", organizationId)) {
            throw new IdsasServiceException(ErrorMessages.shakCodeContainsSpecialCharacter(), FaultCode.CREATE_ORG_BLURRING_ERROR, true);
        }

        OrgBlurringEntity orgBlurringEntity = new OrgBlurringEntity();
        orgBlurringEntity.setCreatorCvr(creatorCvr);
        orgBlurringEntity.setOrganisationId(organizationId);
        orgBlurringEntity.setOrganisationIdType(organizationIdType.name());
        orgBlurringEntity.setValidFrom(Instant.now());
        orgBlurringEntity.setValidTo(null); // Ingen udløbsdato

        // Der må ikke allerede være en registrering
        if (orgBlurringDao.orgBlurringExists(organizationId, organizationIdType.name())) {
            throw new IdsasServiceException(ErrorMessages.orgBlurringAlreadyExists(), FaultCode.CREATE_ORG_BLURRING_ERROR, true);
        }

        // Blurring oprettes via valgte storage strategi
        createStorage.apply(orgBlurringEntity);
    }

    /**
     * Returnerer listen af registrerede afdelinger hørende til CVR nummeret i SOSI ID-kortet.
     */
    public List<OrganizationIdDto> listOrgBlurringsForCVR(String creatorCvr, IdsasActor.IdsasActorType actorType) throws IdsasServiceException {
        logger.trace("List org blurrings for CVR");

        // Tjek at det er en SYSTEM-bruger der kalder
        if (!actorType.equals(IdsasActor.IdsasActorType.SYSTEM)) {
            throw new IdsasServiceException(ErrorMessages.userTypeNotSupported(actorType.name()),
                    FaultCode.LIST_ORG_BLURRING_ERROR, true);
        }

        List<OrgBlurringEntity> entities = orgBlurringDao.getActiveOrgBlurrings(creatorCvr);
        return entities.stream()
                .map(entity -> new OrganizationIdDto(
                        entity.getOrganisationId(),
                        OrgIdType.valueOf(entity.getOrganisationIdType())))
                .collect(Collectors.toList());
    }

    /**
     * Returnerer listen af alle aktive afdelingssløringer. Skal bruges af tjenester, der skal effektuere afdelingssløringer.
     */
    public List<OrganizationIdDto> listAllActiveOrgBlurrings(String clientSystemName, IdsasActor.IdsasActorType actorType) throws IdsasServiceException {
        logger.trace("List all active org blurrings");

        // Tjek at det er en SYSTEM-bruger der kalder
        if (!actorType.equals(IdsasActor.IdsasActorType.SYSTEM)) {
            throw new IdsasServiceException(ErrorMessages.userTypeNotSupported(actorType.name()),
                    FaultCode.LIST_ORG_BLURRING_ERROR, true);
        }

        // Validering af, at systembrugeren har et clientName med den korrekte værdi
        if (!allowedClientSystemNames.contains(clientSystemName)) {
            String names = String.join(", ", allowedClientSystemNames);
            throw new IdsasServiceException(ErrorMessages.clientSystemNameError(
                    clientSystemName, names), FaultCode.LOOKUP_ERROR, true);
        }

        List<OrgBlurringEntity> entities = orgBlurringDao.getActiveOrgBlurrings();
        return entities.stream()
                .map(entity -> new OrganizationIdDto(
                        entity.getOrganisationId(),
                        OrgIdType.valueOf(entity.getOrganisationIdType())))
                .collect(Collectors.toList());
    }

    /**
     * Inaktiverer en afdelingssløring per 'nu' (sætter validTo til 'nu').<br/>
     * Funktionen fejler hvis man forsøger at fjerne en ikke eksisterende afdelingssløring eller hvis man forsøger at fjerne en afdelingssløring, der ikke hører til kalderens CVR-nummer.
     */
    public void removeOrgBlurring(String creatorCvr, String organizationId, OrgIdType organizationIdType) throws IdsasServiceException {
        logger.trace("Remove org blurring");

        // Hent den aktive slørring
        List<OrgBlurringEntity> orgBlurringEntities = orgBlurringDao.getActiveOrgBlurrings(organizationId, organizationIdType.name());
        if (orgBlurringEntities.isEmpty()) {
            throw new IdsasServiceException(ErrorMessages.orgBlurringDoesNotExist(), FaultCode.REMOVE_ORG_BLURRING_ERROR, false);
        }

        // Tjek at der kun er én aktiv slørring for organisationen
        if (orgBlurringEntities.size() > 1) {
            throw new IdsasServiceException(ErrorMessages.multipleOrgBlurrings(), FaultCode.REMOVE_ORG_BLURRING_ERROR, false);
        }

        // Tjek at creatorCvr matcher på registreringen
        if (!orgBlurringEntities.get(0).getCreatorCvr().equals(creatorCvr)) {
            throw new IdsasServiceException(ErrorMessages.orgBlurringNotCreatedByOrg(), FaultCode.REMOVE_ORG_BLURRING_ERROR, false);
        }

        // Sløringen annulleres via valgte storage strategi
        removeStorage.apply(orgBlurringEntities.get(0));
    }

}
