package dk.nsp.idsas.kafka;

import static org.mockito.Mockito.*;
import static org.junit.Assert.*;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;
import dk.nsp.idsas.dao.BlurringDao;
import dk.nsp.idsas.dao.OrgBlurringDao;
import dk.nsp.idsas.dao.entity.BlurringEntity;
import dk.nsp.idsas.dao.entity.OrgBlurringEntity;
import dk.nsp.idsas.dto.KafkaMsgType;
import dk.nsp.idsas.service.IdsasServiceException;
import dk.nsp.idsas.util.health.KafkaConsumeStatus;
import org.apache.kafka.clients.consumer.Consumer;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.header.internals.RecordHeader;
import org.apache.kafka.common.errors.WakeupException;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;

import java.nio.charset.StandardCharsets;
import java.time.Instant;
import java.util.*;

public class KafkaConsumerServiceTest {

    private KafkaConsumerService kafkaConsumerService;
    private ObjectMapper mapper;

    @Mock
    private Consumer<String, String> mockConsumer;

    @Mock
    private BlurringDao mockBlurringDao;

    @Mock
    private OrgBlurringDao mockOrgBlurringDao;

    private final long pollTimeout = 1000L;

    @Before
    public void setup() {
        MockitoAnnotations.openMocks(this);

        mapper = new ObjectMapper();
        mapper.disable(SerializationFeature.WRITE_DATES_AS_TIMESTAMPS);
        mapper.registerModule(new JavaTimeModule());

        kafkaConsumerService = new KafkaConsumerService(mockConsumer, pollTimeout, mockBlurringDao, mockOrgBlurringDao);
    }

    @Test
    public void testConsume_CreateBlurringEntity() throws Exception {
        // Arrange
        BlurringEntity entity = new BlurringEntity("cc04a74b4cf982b2930e7a328471acd18b0e24442705ad9256cc8c053434508c",
                "CPR", "33257872", "CVR", Instant.parse("2024-09-13T08:30:36.013Z"));
        KafkaMessage message = new KafkaMessage(
                UUID.randomUUID().toString(),
                KafkaMsgType.CREATE_BLURRING_ENTITY,
                mapper.writeValueAsString(entity));
        String jsonMessage = mapper.writeValueAsString(message);

        Map<TopicPartition, List<ConsumerRecord<String, String>>> records = new LinkedHashMap<>();
        ConsumerRecord<String, String> record1 = new ConsumerRecord<>("topic", 0, 0L, null, jsonMessage);
        record1.headers().add(new RecordHeader("type", KafkaMsgType.CREATE_BLURRING_ENTITY.getValue().getBytes(StandardCharsets.UTF_8)));
        records.put(new TopicPartition("topic", 0), Collections.singletonList(record1));

        when(mockConsumer.poll(pollTimeout)).thenReturn(new ConsumerRecords<>(records));

        // Act
        kafkaConsumerService.consume();

        // Assert
        ArgumentCaptor<BlurringEntity> captor = ArgumentCaptor.forClass(BlurringEntity.class);
        verify(mockBlurringDao).insert(captor.capture());
        BlurringEntity capturedEntity = captor.getValue();
        assertNotNull(capturedEntity);
        assertEquals(entity, capturedEntity);
                assertEquals(KafkaConsumeStatus.FINISHED, kafkaConsumerService.getStatus());
        verify(mockConsumer).commitSync();
    }

    @Test
    public void testConsume_CreateOrgBlurringEntity() throws Exception {
        // Arrange
        OrgBlurringEntity entity = new OrgBlurringEntity("cc04a74b4cf982b2930e7a328471acd18b0e24442705ad9256cc8c053434508c",
                "CVR", "33257872", Instant.parse("2024-09-13T08:30:37.013Z"), Instant.parse("2024-09-13T08:30:36.013Z"));
        KafkaMessage message = new KafkaMessage(
                UUID.randomUUID().toString(),
                KafkaMsgType.CREATE_ORG_BLURRING_ENTITY,
                mapper.writeValueAsString(entity));
        String jsonMessage = mapper.writeValueAsString(message);

        Map<TopicPartition, List<ConsumerRecord<String, String>>> records = new LinkedHashMap<>();
        ConsumerRecord<String, String> record1 = new ConsumerRecord<>("topic", 0, 0L, null, jsonMessage);
        record1.headers().add(new RecordHeader("type", KafkaMsgType.CREATE_ORG_BLURRING_ENTITY.getValue().getBytes(StandardCharsets.UTF_8)));
        records.put(new TopicPartition("topic", 0), Collections.singletonList(record1));

        when(mockConsumer.poll(pollTimeout)).thenReturn(new ConsumerRecords<>(records));

        // Act
        kafkaConsumerService.consume();

        // Assert
        ArgumentCaptor<OrgBlurringEntity> captor = ArgumentCaptor.forClass(OrgBlurringEntity.class);
        verify(mockOrgBlurringDao).insert(captor.capture());
        OrgBlurringEntity capturedEntity = captor.getValue();
        assertNotNull(capturedEntity);
        assertEquals(entity, capturedEntity);
        assertEquals(KafkaConsumeStatus.FINISHED, kafkaConsumerService.getStatus());
        verify(mockConsumer).commitSync();
    }

    @Test
    public void testConsume_RemoveOrgBlurringEntity() throws Exception {
        // Arrange
        OrgBlurringEntity entity = new OrgBlurringEntity("cc04a74b4cf982b2930e7a328471acd18b0e24442705ad9256cc8c053434508c",
                "CVR", "33257872", Instant.parse("2024-09-13T08:30:37.013Z"), Instant.parse("2024-09-13T08:30:36.013Z"));
        KafkaMessage message = new KafkaMessage(
                UUID.randomUUID().toString(),
                KafkaMsgType.REMOVE_ORG_BLURRING_ENTITY,
                mapper.writeValueAsString(entity));
        String jsonMessage = mapper.writeValueAsString(message);

        Map<TopicPartition, List<ConsumerRecord<String, String>>> records = new LinkedHashMap<>();
        ConsumerRecord<String, String> record1 = new ConsumerRecord<>("topic", 0, 0L, null, jsonMessage);
        record1.headers().add(new RecordHeader("type", KafkaMsgType.REMOVE_ORG_BLURRING_ENTITY.getValue().getBytes(StandardCharsets.UTF_8)));
        records.put(new TopicPartition("topic", 0), Collections.singletonList(record1));

        when(mockConsumer.poll(pollTimeout)).thenReturn(new ConsumerRecords<>(records));

        // Act
        kafkaConsumerService.consume();

        // Assert
        verify(mockOrgBlurringDao).removeOrgBlurring(entity.getOrganisationId(), entity.getOrganisationIdType());
        assertEquals(KafkaConsumeStatus.FINISHED, kafkaConsumerService.getStatus());
        verify(mockConsumer).commitSync();
    }

    @Test
    public void testConsume_EmptyRecords() throws IdsasServiceException {
        // Arrange
        when(mockConsumer.poll(pollTimeout)).thenReturn(new ConsumerRecords<>(Collections.emptyMap()));

        // Act
        kafkaConsumerService.consume();

        // Assert
        assertEquals(KafkaConsumeStatus.FINISHED, kafkaConsumerService.getStatus());
        verify(mockConsumer, never()).commitSync();
    }


    @Test
    public void testConsume_JSONParseException() {
        // Arrange
        Map<TopicPartition, List<ConsumerRecord<String, String>>> records = new LinkedHashMap<>();
        ConsumerRecord<String, String> record1 = new ConsumerRecord<>("topic", 0, 0L, null, "invalid json");
        record1.headers().add(new RecordHeader("type", KafkaMsgType.CREATE_BLURRING_ENTITY.getValue().getBytes(StandardCharsets.UTF_8)));
        records.put(new TopicPartition("topic", 0), Collections.singletonList(record1));

        when(mockConsumer.poll(pollTimeout)).thenReturn(new ConsumerRecords<>(records));

        // Act & Assert
        try {
            kafkaConsumerService.consume();
            fail("Expected IdsasServiceException not thrown");
        } catch (IdsasServiceException e) {
            assertEquals("Error parsing JSON message", e.getMessage());
            assertEquals(KafkaConsumeStatus.ERROR, kafkaConsumerService.getStatus());
        }
    }

    @Test
    public void testConsume_WakeupException() {
        // Arrange
        when(mockConsumer.poll(pollTimeout)).thenThrow(new WakeupException());

        // Act & Assert
        try {
            kafkaConsumerService.consume();
            fail("Expected IdsasServiceException not thrown");
        } catch (IdsasServiceException e) {
            assertEquals("Wakeup called on Kafka consumer", e.getMessage());
            assertEquals(KafkaConsumeStatus.ERROR, kafkaConsumerService.getStatus());
        }
    }

    @Test
    public void testStopConsume() {
        // Act
        kafkaConsumerService.stopConsume();

        // Assert
        verify(mockConsumer).wakeup();
    }

    @Test
    public void testConsume_auditLogHappens() throws JsonProcessingException, IdsasServiceException {
        // Arrange
        BlurringEntity entity = new BlurringEntity("cc04a74b4cf982b2930e7a328471acd18b0e24442705ad9256cc8c053434508c",
                "CPR", "33257872", "CVR", Instant.parse("2024-09-13T08:30:36.013Z"));
        KafkaMessage message = new KafkaMessage(
                UUID.randomUUID().toString(),
                KafkaMsgType.CREATE_BLURRING_ENTITY,
                mapper.writeValueAsString(entity));
        String jsonMessage = mapper.writeValueAsString(message);

        Map<TopicPartition, List<ConsumerRecord<String, String>>> records = new LinkedHashMap<>();
        ConsumerRecord<String, String> record1 = new ConsumerRecord<>("topic", 0, 0L, null, jsonMessage);
        record1.headers().add(new RecordHeader("type", KafkaMsgType.CREATE_BLURRING_ENTITY.getValue().getBytes(StandardCharsets.UTF_8)));
        records.put(new TopicPartition("topic", 0), Collections.singletonList(record1));

        when(mockConsumer.poll(pollTimeout)).thenReturn(new ConsumerRecords<>(records));

        kafkaConsumerService = new KafkaConsumerService(mockConsumer, pollTimeout, mockBlurringDao, mockOrgBlurringDao) {
            @Override
            protected void auditLog(String k, String t, Object m) {
                // Assert
                assertEquals(message.getKey(), k);
                assertEquals(message.getType().getValue(), t);
                assertTrue(m instanceof BlurringEntity);
            }
        };

        // Act
        kafkaConsumerService.consume();
    }
}