package dk.nsp.idsas.kafka;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import dk.nsp.idsas.dao.entity.BlurringEntity;
import dk.nsp.idsas.dao.entity.OrgBlurringEntity;
import dk.nsp.idsas.service.IdsasServiceException;
import org.apache.kafka.clients.producer.Producer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.clients.producer.RecordMetadata;
import org.apache.kafka.common.TopicPartition;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;

import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import static org.junit.Assert.assertEquals;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;


public class KafkaProducerServiceTest {

    private KafkaProducerService kafkaProducerService;

    @Mock
    private Producer<String, String> kafkaProducer;

    private ObjectMapper objectMapper;

    @Before
    public void setup() {
        MockitoAnnotations.openMocks(this);

        objectMapper = new ObjectMapper();

        TopicPartition topicPartition = new TopicPartition("topic", 0);
        RecordMetadata recordMetadata = new RecordMetadata(topicPartition, 0L, 0, 0L, 0L, 0, 0);
        Future<RecordMetadata> recordMetadataFuture = new Future<RecordMetadata>() {
            @Override
            public boolean cancel(boolean mayInterruptIfRunning) {
                return false;
            }

            @Override
            public boolean isCancelled() {
                return false;
            }

            @Override
            public boolean isDone() {
                return true;
            }

            @Override
            public RecordMetadata get(){
                return recordMetadata;
            }

            @Override
            public RecordMetadata get(long timeout, TimeUnit unit) {
                return recordMetadata;
            }
        };

        when(kafkaProducer.send(any())).thenReturn(recordMetadataFuture);
    }

    @Test
    public void sendCreateBlurringMessage_shouldSendCreateBlurringMessage() throws IdsasServiceException, JsonProcessingException {
        // Arrange
        kafkaProducerService = new KafkaProducerService("topic", kafkaProducer);
        BlurringEntity entity = new BlurringEntity("patientIdHash", "patientIdType", "organisationId", "organisationIdType", null);

        // Act
        kafkaProducerService.sendCreateBlurringMessage(entity);

        // Assert
        ArgumentCaptor<ProducerRecord<String, String>> captor = ArgumentCaptor.forClass(ProducerRecord.class);
        verify(kafkaProducer).send(captor.capture());
        ProducerRecord<String, String> capturedRecord = captor.getValue();
        KafkaMessage capturedMessage = objectMapper.readValue(capturedRecord.value(), KafkaMessage.class);
        BlurringEntity capturedEntity = objectMapper.readValue(capturedMessage.getMessage(), BlurringEntity.class);
        assertEquals(entity, capturedEntity);
    }

    @Test
    public void sendCreateOrgBlurringMessage_shouldSendCreateOrgBlurringMessage() throws IdsasServiceException, JsonProcessingException {
        // Arrange
        kafkaProducerService = new KafkaProducerService("topic", kafkaProducer);
        OrgBlurringEntity entity = new OrgBlurringEntity("patientIdHash", "organisationId", "organisationIdType", null, null);

        // Act
        kafkaProducerService.sendCreateOrgBlurringMessage(entity);

        // Assert
        ArgumentCaptor<ProducerRecord<String, String>> captor = ArgumentCaptor.forClass(ProducerRecord.class);
        verify(kafkaProducer).send(captor.capture());
        ProducerRecord<String, String> capturedRecord = captor.getValue();
        KafkaMessage capturedMessage = objectMapper.readValue(capturedRecord.value(), KafkaMessage.class);
        OrgBlurringEntity capturedEntity = objectMapper.readValue(capturedMessage.getMessage(), OrgBlurringEntity.class);
        assertEquals(entity, capturedEntity);
    }

    @Test
    public void removeOrgBlurringMessage_shouldSendRemoveOrgBlurringMessage() throws IdsasServiceException, JsonProcessingException {
        // Arrange
        kafkaProducerService = new KafkaProducerService("topic", kafkaProducer);
        OrgBlurringEntity entity = new OrgBlurringEntity("patientIdHash", "organisationId", "organisationIdType", null, null);

        // Act
        kafkaProducerService.removeOrgBlurringMessage(entity);

        // Assert
        ArgumentCaptor<ProducerRecord<String, String>> captor = ArgumentCaptor.forClass(ProducerRecord.class);
        verify(kafkaProducer).send(captor.capture());
        ProducerRecord<String, String> capturedRecord = captor.getValue();
        KafkaMessage capturedMessage = objectMapper.readValue(capturedRecord.value(), KafkaMessage.class);
        OrgBlurringEntity capturedEntity = objectMapper.readValue(capturedMessage.getMessage(), OrgBlurringEntity.class);
        assertEquals(entity, capturedEntity);
    }

    @Test
    public void sendMessage_shouldThrowIdsasServiceException() {
        // Arrange
        kafkaProducerService = new KafkaProducerService("topic", kafkaProducer);
        BlurringEntity entity = new BlurringEntity("patientIdHash", "patientIdType", "organisationId", "organisationIdType", null);
        when(kafkaProducer.send(any())).thenThrow(new RuntimeException("Error in Kafka"));

        // Act
        try {
            kafkaProducerService.sendCreateBlurringMessage(entity);
        } catch (IdsasServiceException e) {
            // Assert
            assertEquals("Sløringen kunne ikke gennemføres. Fejl ved afsendelse af Kafka besked.", e.getMessage());
        }
    }

    @Test
    public void sendMessage_auditLogHappens() throws IdsasServiceException {
        // Arrange
        BlurringEntity entity = new BlurringEntity("patientIdHash", "patientIdType", "organisationId", "organisationIdType", null);
        kafkaProducerService = new KafkaProducerService("topic", kafkaProducer) {
            @Override
            protected void auditLog(String key, String type, Object message) {
                assertEquals("CreateBlurringEntity", type);
                assertEquals(entity, message);
            }
        };

        // Act
        kafkaProducerService.sendCreateBlurringMessage(entity);
    }

}
