package dk.nsp.idsas.service;

import dk.nsp.idsas.dao.WhitelistDao;
import dk.nsp.idsas.dao.entity.BlurringEntity;
import dk.nsp.idsas.dto.OrgIdType;
import dk.nsp.idsas.dto.PersonIdType;
import dk.nsp.idsas.integrations.personinformation.PersonInformationClient;
import dk.nsp.idsas.integrations.personinformation.PersonInformationClientException;
import dk.nsp.idsas.security.IdsasActor;
import dk.nsp.idsas.util.AuditLogger;
import dk.nsp.idsas.util.ErrorMessages;
import dk.nsp.idsas.util.WhitelistType;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.Mockito;

import java.time.ZonedDateTime;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertThrows;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.Mockito.mock;
import static org.mockito.MockitoAnnotations.openMocks;

public class BlurringServiceTest {
    @Mock
    private BlurringStorage<BlurringEntity> blurringStorage;

    private WhitelistDao whitelistDao;
    private AuditLogger auditLogger;
    private PersonInformationClient personInformationClient;
    private BlurringService blurringService;
    private final int maxExpiryDays = 90;

    @Before
    public void setUp() {
        openMocks(this);

        whitelistDao = mock(WhitelistDao.class);
        auditLogger = mock(AuditLogger.class);
        personInformationClient = mock(PersonInformationClient.class);
        blurringService = new BlurringService(whitelistDao, auditLogger, personInformationClient, blurringStorage, maxExpiryDays);
    }

    @Test
    public void createBlurring_WhenCvrIsNotWhitelisted_ThrowsException() {
        // Arrange
        IdsasActor actor = IdsasActor.createSystem("5678", "CVR", "TestSystem", "");
        when(whitelistDao.isWhitelisted(actor.getOrganizationId(), WhitelistType.BLURRING)).thenReturn(false);

        // Act
        IdsasServiceException e = assertThrows(IdsasServiceException.class,
                () -> blurringService.createBlurring(actor.getOrganizationId(), OrgIdType.CVR, "", PersonIdType.CPR, ZonedDateTime.now()));

        // Assert
        assertEquals(ErrorMessages.cvrWhitelistError(actor.getOrganizationId(), WhitelistType.BLURRING), e.getMessage());
    }

    @Test
    public void createBlurring_WhenNoPersonInformationIsFoundForCpr_ThrowsException() throws PersonInformationClientException {
        // Arrange
        IdsasActor actor = IdsasActor.createSystem("5678", "CVR", "TestSystem", "");
        String cpr = "1111111111";
        when(whitelistDao.isWhitelisted(actor.getOrganizationId(), WhitelistType.BLURRING)).thenReturn(true);
        when(personInformationClient.personExists(cpr)).thenReturn(false);

        // Act
        IdsasServiceException e = assertThrows(IdsasServiceException.class,
                () -> blurringService.createBlurring(actor.getOrganizationId(), OrgIdType.CVR, cpr, PersonIdType.CPR, ZonedDateTime.now()));

        // Assert
        assertEquals(ErrorMessages.cprIsNotValidError(), e.getMessage());
    }

    @Test
    public void createBlurring_WhenAllValidationsSucceed_CreatesBlurringAndAuditLogs() throws PersonInformationClientException, IdsasServiceException {
        // Arrange
        IdsasActor actor = IdsasActor.createSystem("5678", "CVR", "TestSystem", "");
        String patientId = "1111111111";
        ZonedDateTime endDateTime = ZonedDateTime.now();
        when(whitelistDao.isWhitelisted(actor.getOrganizationId(), WhitelistType.BLURRING)).thenReturn(true);
        when(personInformationClient.personExists(patientId)).thenReturn(true);
        Mockito.when(whitelistDao.isWhitelisted(actor.getOrganizationId(), WhitelistType.BLURRING)).thenReturn(true);
        Mockito.when(personInformationClient.personExists(patientId)).thenReturn(true);

        // Act
        blurringService.createBlurring(actor.getOrganizationId(), OrgIdType.CVR, patientId, PersonIdType.CPR, endDateTime);

        // Assert
        verify(auditLogger).logCreateBlurring(patientId, "CPR", actor.getOrganizationId(), OrgIdType.CVR.name(), endDateTime.toInstant());
        ArgumentCaptor<BlurringEntity> captor = ArgumentCaptor.forClass(BlurringEntity.class);
        verify(blurringStorage).apply(captor.capture());
        BlurringEntity blurringEntity = captor.getValue();
        assertEquals(actor.getOrganizationId(), blurringEntity.getOrganisationId());
        assertEquals(OrgIdType.CVR.name(), blurringEntity.getOrganisationIdType());
    }

    @Test
    public void createBlurring_WhenExpiryDateIsExceeded_ThrowsException() throws PersonInformationClientException {
        // Arrange
        IdsasActor actor = IdsasActor.createSystem("5678", "CVR", "TestSystem", "");
        ZonedDateTime exceededDate = ZonedDateTime.now().plusYears(2).plusDays(1);
        String patientId = "1111111111";
        when(whitelistDao.isWhitelisted(actor.getOrganizationId(), WhitelistType.BLURRING)).thenReturn(true);
        when(personInformationClient.personExists(patientId)).thenReturn(true);

        // Act
        IdsasServiceException e = assertThrows(IdsasServiceException.class,
                () -> blurringService.createBlurring(actor.getOrganizationId(), OrgIdType.CVR, patientId, PersonIdType.CPR, exceededDate));

        // Assert
        assertEquals(ErrorMessages.maxExpiryDateExceeded(maxExpiryDays), e.getMessage());
    }

}
