package dk.nsp.idsas.service;

import dk.nsp.idsas.dao.BlurringDao;
import dk.nsp.idsas.integrations.personinformation.PersonInformationClient;
import dk.nsp.idsas.integrations.personinformation.PersonInformationClientException;
import dk.nsp.idsas.util.health.CleanupJobStatus;
import org.apache.http.client.HttpClient;
import org.junit.Before;
import org.junit.Test;

import java.time.Duration;
import java.time.Instant;
import java.time.Period;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;

import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class CleanupDeceasedServiceTest {
    private BlurringDao blurringDao;
    private PersonInformationClient personInformationClient;

    @Before
    public void setUp() throws PersonInformationClientException {
        blurringDao = mock(BlurringDao.class);
        personInformationClient = new PersonInformationClient(
                mock(HttpClient.class),
                "http://test1-cnsp.ekstern-test.nspop.dk:8080/stamdata-personinformation",
                Duration.ofMinutes(5)) {

            @Override
            public List<String> getDeceased(List<String> listOfCpr, Instant deceasedSince)  {
                return listOfCpr.stream().filter(cpr -> Arrays.asList("1", "2", "3", "4", "8").contains(cpr)).collect(Collectors.toList());
            }
        };
    }

    @Test
    public void doCleanup_FiveDeadPeopleAndTwoBatches_AllDeleted() throws IdsasServiceException {
        // Arrange
        CleanupDeceasedService cleanupDeceasedService = new CleanupDeceasedService(blurringDao, personInformationClient, Duration.ofMinutes(5), 5, Period.ofYears(1));
        when(blurringDao.getAllBlurrings()).thenReturn(new LinkedList<>(Arrays.asList("1","2","3","4","5","6","7","8","9","10")));

        // Act
        cleanupDeceasedService.cleanUpDeceasedBlurrings();

        // Assert
        CleanupJobStatus status = cleanupDeceasedService.getStatus();
        assertEquals(5, status.rowsDeleted);
        assertEquals(0, status.rowsLeft);
        assertEquals(CleanupJobStatus.ExecutionStatus.FINISHED, status.finished);
    }

    @Test
    public void doCleanup_NoDeadPeople_NothingDeleted() throws IdsasServiceException {
        // Arrange
        CleanupDeceasedService cleanupDeceasedService = new CleanupDeceasedService(blurringDao, personInformationClient, Duration.ofMinutes(5), 5, Period.ofYears(1));
        when(blurringDao.getAllBlurrings()).thenReturn(new LinkedList<>(Arrays.asList("5","6","7","9","10")));

        // Act
        cleanupDeceasedService.cleanUpDeceasedBlurrings();

        // Assert
        CleanupJobStatus status = cleanupDeceasedService.getStatus();
        assertEquals(0, status.rowsDeleted);
        assertEquals(0, status.rowsLeft);
        assertEquals(CleanupJobStatus.ExecutionStatus.FINISHED, status.finished);
    }

    @Test
    public void doCleanup_TimeoutAfterFirstBatch_FiveRowsLeft() throws IdsasServiceException {
        // Arrange
        CleanupDeceasedService cleanupDeceasedService = new CleanupDeceasedService(blurringDao, personInformationClient, Duration.ofMillis(1), 5, Period.ofYears(1)) {
            private boolean first = true;
            @Override
            protected boolean belowTimeLimit(long startMillis) {
                // First time we return true, second time we return false
                if (first) {
                    first = false;
                    return true;
                }
                return false;
            }
        };
        when(blurringDao.getAllBlurrings()).thenReturn(new LinkedList<>(Arrays.asList("1","2","3","4","5","6","7","8","9","10")));

        // Act
        cleanupDeceasedService.cleanUpDeceasedBlurrings();

        // Assert
        CleanupJobStatus status = cleanupDeceasedService.getStatus();
        assertEquals(4, status.rowsDeleted);
        assertEquals(5, status.rowsLeft);
        assertEquals(CleanupJobStatus.ExecutionStatus.NOT_FINISHED, status.finished);
    }

}
