package dk.nsp.idsas.service;

import dk.nsp.idsas.dao.OrgBlurringDao;
import dk.nsp.idsas.dao.WhitelistDao;
import dk.nsp.idsas.dao.entity.OrgBlurringEntity;
import dk.nsp.idsas.dto.OrgIdType;
import dk.nsp.idsas.dto.OrganizationIdDto;
import dk.nsp.idsas.integrations.sores.SoresClient;
import dk.nsp.idsas.integrations.sores.SoresClientException;
import dk.nsp.idsas.security.IdsasActor;
import dk.nsp.idsas.util.ErrorMessages;
import dk.nsp.idsas.util.FaultCode;
import dk.nsp.idsas.util.WhitelistType;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;

import java.util.ArrayList;
import java.util.List;

import static org.junit.Assert.*;
import static org.mockito.Mockito.*;

public class OrgBlurringServiceTest {
    private SoresClient soresClient;
    private OrgBlurringDao orgBlurringDao;
    private WhitelistDao whiteListDao;
    private final String[] allowedClientSystemNames = new String[]{"system1", "system2"};

    @Mock
    private BlurringStorage<OrgBlurringEntity> createStorage;
    @Mock
    private BlurringStorage<OrgBlurringEntity> removeStorage;

    @Before
    public void setUp() {
        MockitoAnnotations.openMocks(this);
        soresClient = Mockito.mock(SoresClient.class);
        orgBlurringDao = Mockito.mock(OrgBlurringDao.class);
        whiteListDao = Mockito.mock(WhitelistDao.class);
    }

    @Test
    public void createOrgBlurring_sorExistsRegistrationDoesNot_noErrors() throws IdsasServiceException, SoresClientException {
        // Act
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames, createStorage, removeStorage);
        String sor = "12345678";
        String cvr = "123";

        when(soresClient.getCvrNumberBySor(sor)).thenReturn(cvr);
        when(orgBlurringDao.orgBlurringExists("123", "SOR")).thenReturn(false);
        when(whiteListDao.isWhitelisted(cvr, WhitelistType.BLURRING)).thenReturn(true);

        // Act
        orgBlurringService.createOrgBlurring(cvr, IdsasActor.IdsasActorType.SYSTEM, sor, OrgIdType.SOR);

        // Assert
        ArgumentCaptor<OrgBlurringEntity> argumentCaptor = ArgumentCaptor.forClass(OrgBlurringEntity.class);
        verify(createStorage).apply(argumentCaptor.capture());
        assertEquals(cvr, argumentCaptor.getValue().getCreatorCvr());
        assertEquals(sor, argumentCaptor.getValue().getOrganisationId());
    }

    @Test
    public void createOrgBlurring_registrationAlreadyExists_shouldFail() throws SoresClientException, IdsasServiceException {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames, createStorage, removeStorage);
        final String sor = "123";
        final String cvr = "12345678";

        when(orgBlurringDao.orgBlurringExists(sor, "SOR")).thenReturn(true);
        when(whiteListDao.isWhitelisted(cvr, WhitelistType.BLURRING)).thenReturn(true);
        when(soresClient.getCvrNumberBySor(sor)).thenReturn(cvr);

        // Act
        IdsasServiceException exception = assertThrows(IdsasServiceException.class,
                () -> orgBlurringService.createOrgBlurring(cvr, IdsasActor.IdsasActorType.SYSTEM, sor, OrgIdType.SOR));

        // Assert
        assertEquals(ErrorMessages.orgBlurringAlreadyExists(), exception.getMessage());
    }

    @Test
    public void createOrgBlurring_notWhiteListed_shouldFail() {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames, createStorage, removeStorage);
        when(whiteListDao.isWhitelisted("12345678", WhitelistType.BLURRING)).thenReturn(false);

        // Act
        IdsasServiceException exception = assertThrows(IdsasServiceException.class,
                () -> orgBlurringService.createOrgBlurring("12345678", IdsasActor.IdsasActorType.SYSTEM, "123", OrgIdType.SOR));

        // Assert
        assertEquals(ErrorMessages.cvrWhitelistError("12345678", WhitelistType.BLURRING), exception.getMessage());
    }

    @Test
    public void createOrgBlurring_usingSpecialCharacterInShakCode_shouldFail() {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames, createStorage, removeStorage);
        when(whiteListDao.isWhitelisted("12345678", WhitelistType.BLURRING)).thenReturn(true);

        // Act
        IdsasServiceException exception = assertThrows(IdsasServiceException.class,
                () -> orgBlurringService.createOrgBlurring("12345678", IdsasActor.IdsasActorType.SYSTEM, "\"!123#", OrgIdType.SHAK));

        // Assert
        assertEquals(ErrorMessages.shakCodeContainsSpecialCharacter(), exception.getMessage());
    }

    @Test
    public void createOrgBlurring_shakContainsUpperCaseLetters_noProblem() throws IdsasServiceException {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames, createStorage, removeStorage);
        when(whiteListDao.isWhitelisted("12345678", WhitelistType.BLURRING)).thenReturn(true);

        // Act
        orgBlurringService.createOrgBlurring("12345678", IdsasActor.IdsasActorType.SYSTEM, "ABC123", OrgIdType.SHAK);

        // No error
    }

    @Test
    public void createOrgBlurring_notSorOrShak_shouldFail() {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames, createStorage, removeStorage);
        when(whiteListDao.isWhitelisted("12345678", WhitelistType.BLURRING)).thenReturn(true);

        // Act
        IdsasServiceException exception = assertThrows(IdsasServiceException.class,
                () -> orgBlurringService.createOrgBlurring("12345678", IdsasActor.IdsasActorType.SYSTEM, "123", OrgIdType.CVR));

        // Assert
        assertEquals(ErrorMessages.orgIdTypeNotSupported(OrgIdType.CVR.name()), exception.getMessage());
    }

    @Test
    public void listOrgBlurringsForCVR_shouldReturnList() throws IdsasServiceException {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames);

        OrgBlurringEntity orgBlurringEntity = new OrgBlurringEntity();
        orgBlurringEntity.setOrganisationId("123");
        orgBlurringEntity.setOrganisationIdType("SOR");
        List<OrgBlurringEntity> orgBlurringEntities = new ArrayList<>();
        orgBlurringEntities.add(orgBlurringEntity);
        when(orgBlurringDao.getActiveOrgBlurrings("123")).thenReturn(orgBlurringEntities);

        // Act
        List<OrganizationIdDto> result = orgBlurringService.listOrgBlurringsForCVR("123", IdsasActor.IdsasActorType.SYSTEM);

        // Assert
        assertEquals(1, result.size());
        assertEquals("123", result.get(0).id);
    }

    @Test
    public void listOrgBlurringsForCVR_shouldReturnEmptyList() throws IdsasServiceException {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames);
        when(orgBlurringDao.getActiveOrgBlurrings("123")).thenReturn(new ArrayList<>());

        // Act
        List<OrganizationIdDto> result = orgBlurringService.listOrgBlurringsForCVR("123", IdsasActor.IdsasActorType.SYSTEM);

        // Assert
        assertEquals(0, result.size());
    }

    @Test
    public void listOrgBlurringsForCVR_shouldThrowException() throws IdsasServiceException {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames);
        String errMsg = "fejl";
        when(orgBlurringDao.getActiveOrgBlurrings("123")).thenThrow(new IdsasServiceException(errMsg, FaultCode.DAO_ACCESS_ERROR, false));

        // Act
        IdsasServiceException exception = assertThrows(IdsasServiceException.class,
                () -> orgBlurringService.listOrgBlurringsForCVR("123", IdsasActor.IdsasActorType.SYSTEM));

        // Assert
        assertEquals(errMsg, exception.getMessage());
    }

    @Test
    public void listAllActiveOrgBlurrings_shouldReturnList() throws IdsasServiceException {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames);
        OrgBlurringEntity orgBlurringEntity = new OrgBlurringEntity();
        orgBlurringEntity.setOrganisationId("123");
        orgBlurringEntity.setOrganisationIdType("SOR");
        List<OrgBlurringEntity> orgBlurringEntities = new ArrayList<>();
        orgBlurringEntities.add(orgBlurringEntity);
        when(orgBlurringDao.getActiveOrgBlurrings()).thenReturn(orgBlurringEntities);

        // Act
        List<OrganizationIdDto> result = orgBlurringService.listAllActiveOrgBlurrings("system1", IdsasActor.IdsasActorType.SYSTEM);

        // Assert
        assertEquals(1, result.size());
        assertEquals("123", result.get(0).id);
    }

    @Test
    public void listAllActiveOrgBlurrings_shouldReturnEmptyList() throws IdsasServiceException {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames);
        when(orgBlurringDao.getActiveOrgBlurrings()).thenReturn(new ArrayList<>());

        // Act
        List<OrganizationIdDto> result = orgBlurringService.listAllActiveOrgBlurrings("system1", IdsasActor.IdsasActorType.SYSTEM);

        // Assert
        assertEquals(0, result.size());
    }

    @Test
    public void listAllActiveOrgBlurrings_shouldThrowException() throws IdsasServiceException {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames);
        String errMsg = "fejl";
        when(orgBlurringDao.getActiveOrgBlurrings()).thenThrow(new IdsasServiceException(errMsg, FaultCode.DAO_ACCESS_ERROR, false));

        // Act
        IdsasServiceException exception = assertThrows(IdsasServiceException.class,
                () -> orgBlurringService.listAllActiveOrgBlurrings("system1", IdsasActor.IdsasActorType.SYSTEM));

        // Assert
        assertEquals(errMsg, exception.getMessage());
    }

    @Test
    public void listAllActiveOrgBlurrings_notAllowedClientSystemName_shouldThrowException() {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames);

        // Act
        IdsasServiceException exception = assertThrows(IdsasServiceException.class,
                () -> orgBlurringService.listAllActiveOrgBlurrings("system3", IdsasActor.IdsasActorType.SYSTEM));

        // Assert
        assertTrue(exception.getMessage().contains("system1"));
        assertTrue(exception.getMessage().contains("system2"));
        assertTrue(exception.getMessage().contains("system3"));
    }

    @Test
    public void removeOrgBlurring_shouldRemove() throws IdsasServiceException {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames, createStorage, removeStorage);

        String cvr = "123";
        String sor = "12345678";
        OrgBlurringEntity orgBlurringEntity = new OrgBlurringEntity();
        orgBlurringEntity.setOrganisationId(sor);
        orgBlurringEntity.setOrganisationIdType("SOR");
        orgBlurringEntity.setCreatorCvr(cvr);
        List<OrgBlurringEntity> orgBlurringEntities = new ArrayList<>();
        orgBlurringEntities.add(orgBlurringEntity);
        when(orgBlurringDao.getActiveOrgBlurrings(sor, "SOR")).thenReturn(orgBlurringEntities);

        // Act
        orgBlurringService.removeOrgBlurring(cvr, sor, OrgIdType.SOR);

        // Assert
        ArgumentCaptor<OrgBlurringEntity> argumentCaptor = ArgumentCaptor.forClass(OrgBlurringEntity.class);
        verify(removeStorage).apply(argumentCaptor.capture());
        assertEquals(cvr, argumentCaptor.getValue().getCreatorCvr());
        assertEquals(sor, argumentCaptor.getValue().getOrganisationId());
    }

    @Test
    public void removeOrgBlurring_wrongCreatorCvr_throwsException() throws IdsasServiceException {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames, createStorage, removeStorage);

        String cvr = "123";
        String sor = "12345678";
        OrgBlurringEntity orgBlurringEntity = new OrgBlurringEntity();
        orgBlurringEntity.setOrganisationId(sor);
        orgBlurringEntity.setOrganisationIdType("SOR");
        orgBlurringEntity.setCreatorCvr("456");
        List<OrgBlurringEntity> orgBlurringEntities = new ArrayList<>();
        orgBlurringEntities.add(orgBlurringEntity);
        when(orgBlurringDao.getActiveOrgBlurrings(sor, "SOR")).thenReturn(orgBlurringEntities);

        // Act
        IdsasServiceException exception = assertThrows(IdsasServiceException.class,
                () -> orgBlurringService.removeOrgBlurring(cvr, sor, OrgIdType.SOR));

        // Assert
        assertEquals(ErrorMessages.orgBlurringNotCreatedByOrg(), exception.getMessage());
    }

    @Test
    public void removeOrgBlurring_noOrgBlurring_throwsException() throws IdsasServiceException {
        // Arrange
        OrgBlurringService orgBlurringService = new OrgBlurringService(whiteListDao, orgBlurringDao, soresClient, allowedClientSystemNames, createStorage, removeStorage);

        String cvr = "123";
        String sor = "12345678";
        when(orgBlurringDao.getActiveOrgBlurrings(sor, "SOR")).thenReturn(new ArrayList<>());

        // Act
        IdsasServiceException exception = assertThrows(IdsasServiceException.class,
                () -> orgBlurringService.removeOrgBlurring(cvr, sor, OrgIdType.SOR));

        // Assert
        assertEquals(ErrorMessages.orgBlurringDoesNotExist(), exception.getMessage());
    }
}
